# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

try:
    from collections.abc import Sequence, Iterable
except ImportError:
    from collections import Sequence, Iterable
from enum import Enum
from numbers import Number
from six import string_types, ensure_str, ensure_text
import numpy as np

from .datatype import DataType
from .entity import Entity
from . import util


class OdmlType(Enum):
    """
    OdmlType provides all types currently supported by the odML
    data format. It provides additional information about the
    nature of the values of an odML Property.
    """
    Boolean = 'boolean'
    Int = 'int'
    Float = 'float'
    String = 'string'
    Text = 'text'
    URL = 'url'
    Person = 'person'
    Datetime = 'datetime'
    Date = 'date'
    Time = 'time'

    def __str__(self):
        return self.value

    def compatible(self, value):
        """
        compatible returns True or False depending on whether a
        passed value can be mapped to an OdmlType or not.

        :param value: Any single value
        :return: Boolean
        """
        if (self in (self.String, self.Text, self.URL, self.Person) and
                DataType.get_dtype(value) == DataType.String):
            return True
        elif (self == self.Boolean and
              DataType.get_dtype(value) == DataType.Bool):
            return True
        elif (self == self.Float and
              DataType.get_dtype(value) == DataType.Float):
            return True
        elif self == self.Int and DataType.get_dtype(value) == DataType.Int64:
            return True
        elif (self in (self.Time, self.Date, self.Datetime) and
              DataType.get_dtype(value) == DataType.String):
            # This might need some extra work, treating as String for now, but
            # keeping it separated from other String values.
            return True

        return False

    @classmethod
    def get_odml_type(cls, dtype):
        """
        get_odml_type returns the appropriate OdmlType
        for a handed in nix value DataType.

        :param dtype: nix DataType
        :return: OdmlType
        """

        if dtype == DataType.Float:
            return cls.Float
        elif dtype == DataType.Int64:
            return cls.Int
        elif dtype == DataType.String:
            return cls.String
        elif dtype == DataType.Bool:
            return cls.Boolean

        raise TypeError("No available OdmlType for type '%s'" % dtype)


class Property(Entity):
    """An odML Property"""
    def __init__(self, nixparent, h5dataset):
        super(Property, self).__init__(nixparent, h5dataset)
        self._h5dataset = self._h5group

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, dtype, oid=None):
        util.check_entity_name(name)
        dtype = cls._make_h5_dtype(dtype)

        h5dataset = h5parent.create_dataset(name, shape=(0,), dtype=dtype)
        h5dataset.set_attr("name", name)

        if not util.is_uuid(oid):
            oid = util.create_id()

        h5dataset.set_attr("entity_id", oid)

        newentity = cls(nixparent, h5dataset)
        newentity.force_created_at()
        newentity.force_updated_at()

        return newentity

    @property
    def name(self):
        return self._h5dataset.get_attr("name")

    @property
    def definition(self):
        return self._h5dataset.get_attr("definition")

    @definition.setter
    def definition(self, d):
        util.check_attr_type(d, str)
        self._h5dataset.set_attr("definition", d)

    @property
    def unit(self):
        return self._h5dataset.get_attr("unit")

    @unit.setter
    def unit(self, new):
        if new:
            new = util.units.sanitizer(new)

        if new == "":
            new = None

        util.check_attr_type(new, str)
        self._h5dataset.set_attr("unit", new)

    @property
    def uncertainty(self):
        dataset = self._h5dataset
        x, y, z = dataset._parent.file.attrs["version"]
        if x < 1 or (x == 1 and y < 1) or (x == 1 and y == 1 and z < 1):
            val = self._h5dataset.dataset[:]
            v = val[0]["uncertainty"]
            return v
        return self._h5dataset.get_attr("uncertainty")

    @uncertainty.setter
    def uncertainty(self, uncertainty):
        util.check_attr_type(uncertainty, Number)
        uncertainty = float(uncertainty) if uncertainty is not None else None
        self._h5dataset.set_attr("uncertainty", uncertainty)

    @property
    def reference(self):
        dataset = self._h5dataset
        x, y, z = dataset._parent.file.attrs["version"]
        if x < 1 or (x == 1 and y < 1) or (x == 1 and y == 1 and z < 1):
            val = self._h5dataset.dataset[:]
            v = val[0]["reference"]
            return v
        return self._h5dataset.get_attr("reference")

    @reference.setter
    def reference(self, ref):
        util.check_attr_type(ref, str)
        self._h5dataset.set_attr("reference", ref)

    @property
    def dependency(self):
        return self._h5dataset.get_attr("dependency")

    @dependency.setter
    def dependency(self, dep):
        util.check_attr_type(dep, str)
        self._h5dataset.set_attr("dependency", dep)

    @property
    def dependency_value(self):
        return self._h5dataset.get_attr("dependency_value")

    @dependency_value.setter
    def dependency_value(self, depval):
        util.check_attr_type(depval, str)
        self._h5dataset.set_attr("dependency_value", depval)

    @property
    def value_origin(self):
        return self._h5dataset.get_attr("value_origin")

    @value_origin.setter
    def value_origin(self, origin):
        util.check_attr_type(origin, str)
        self._h5dataset.set_attr("value_origin", origin)

    @property
    def odml_type(self):
        otype = self._h5dataset.get_attr("odml_type")
        if not otype:
            return None

        return OdmlType(otype)

    @odml_type.setter
    def odml_type(self, new_type):
        """
        odml_type can only be set if the handed in new type is a valid
        OdmlType and if it is compatible with the value data type of
        the property.

        :param new_type: OdmlType
        """
        if not isinstance(new_type, OdmlType):
            raise TypeError("'{}' is not a valid odml_type.".format(new_type))

        if not new_type.compatible(self.values[0]):
            raise TypeError("Type '{}' is incompatible "
                            "with property values".format(new_type))

        self._h5dataset.set_attr("odml_type", str(new_type))

    def _read_old_values(self):
        val = self._h5dataset.dataset[:]
        val_tu = tuple()
        for v in val:
            v = v["value"]
            val_tu += (v,)
        return val_tu

    @property
    def values(self):
        dataset = self._h5dataset
        x, y, z = dataset._parent.file.attrs["version"]
        if x < 1 or (x == 1 and y < 1) or (x == 1 and y == 1 and z < 1):
            v = self._read_old_values()
            return v
        if not sum(dataset.shape):
            return tuple()

        data = dataset.read_data()

        def data_to_value(dat):
            if isinstance(dat, bytes):
                dat = ensure_str(dat)  # py2compat
            return dat

        values = tuple(map(data_to_value, data))

        return values

    @values.setter
    def values(self, vals):
        """
        Set the value of the property discarding any previous information.

        :param vals: a single value or list of values.
        """
        # Make sure boolean value 'False' gets through as well...
        if vals is None or (isinstance(vals, (Sequence, Iterable)) and
                            not len(vals)):
            self.delete_values()
            return

        if not isinstance(vals, (Sequence, Iterable)) \
                or isinstance(vals, string_types):
            vals = [vals]

        # Make sure all values are of the same data type
        vtype = self._check_new_value_types(vals)
        if vtype == DataType.String:
            vals = [ensure_text(v) for v in vals]  # py2compat
        self._h5dataset.shape = np.shape(vals)
        data = np.array(vals, dtype=vtype)
        self._h5dataset.write_data(data)

    def extend_values(self, data):
        """
        Extends values to existing data.
        Suitable when new data is nested or original data is long.
        """
        vtype = self._check_new_value_types(data)

        arr = np.array(data, dtype=vtype).flatten('C')
        ds = self._h5dataset
        src_len = len(self.values)
        dlen = len(arr)
        ds.shape = (src_len+dlen,)
        ds.write_data(arr, sl=np.s_[src_len: src_len+dlen])

    def _check_new_value_types(self, data):
        if (isinstance(data, (Sequence, Iterable)) and
                not isinstance(data, string_types)):
            single_val = data[0]
        else:
            single_val = data
            data = [data]

        def check_prop_consistent(vtype):
            # Check if the new data has the same type as the existing property
            # data
            if vtype != self.data_type:
                raise TypeError("New data type '{}' is inconsistent with the "
                                "Property's data type '{}'".format(
                                    vtype, self.data_type))

        def check_new_data_consistent(vtype):
            # Check if each value in the new data has the same type
            for val in data:
                if DataType.get_dtype(val) != vtype:
                    raise TypeError("Array contains inconsistent values. "
                                    "Only values of type '{}' can be "
                                    "assigned".format(vtype))

        if hasattr(data, "dtype"):
            # numpy array: no need to scan values, arrays are consistent but
            # check for 1D
            vtype = data.dtype
            check_prop_consistent(vtype)
        else:
            # Will raise an error, if the data type of the first value is not
            # valid
            vtype = DataType.get_dtype(single_val)
            check_prop_consistent(vtype)
            check_new_data_consistent(vtype)

        return vtype

    @property
    def data_type(self):
        dtype = self._h5dataset.dtype

        if dtype == util.vlen_str_dtype:
            return DataType.String

        return dtype

    def delete_values(self):
        self._h5dataset.shape = (0,)

    @staticmethod
    def _make_h5_dtype(valued_type):
        str_ = util.vlen_str_dtype

        if valued_type == DataType.String:
            valued_type = str_

        return valued_type

    def __str__(self):
        return "{}: {{name = {}}}".format(
            type(self).__name__, self.name
        )

    def __repr__(self):
        return self.__str__()

    def pprint(self, indent=2, max_length=80, current_depth=-1):
        """
        Pretty print method. Method is called in Section.pprint()
        """
        property_spaces = ""
        prefix = ""
        if current_depth >= 0:
            property_spaces = " " * ((current_depth + 2) * indent)
            prefix = "|-"
        if self.unit is None:
            value_string = str(self.values)
        else:
            value_string = "{}{}".format(self.values, self.unit)
        p_len = len(property_spaces) + len(self.name) + len(value_string)
        if p_len >= max_length - 4:
            split_len = int((max_length - len(property_spaces)
                             + len(self.name) - len(prefix))/2)
            str1 = value_string[0: split_len]
            str2 = value_string[-split_len:]
            print(("{}{} {}: {} ... {}".format(property_spaces, prefix,
                                               self.name, str1, str2)))
        else:
            print(("{}{} {}: {}".format(property_spaces, prefix, self.name,
                                        value_string)))



# -*- coding: utf-8 -*-
# Copyright © 2019, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from collections import OrderedDict
from inspect import isclass
import numpy as np
from .exceptions import OutOfBounds
from .entity import Entity
from . import util
from .data_set import DataSet
from .datatype import DataType
from .section import Section
from six import string_types
import csv


class DataFrame(Entity, DataSet):

    def __init__(self, nixparent, h5group):
        super(DataFrame, self).__init__(nixparent, h5group)
        self._sources = None
        self._columns = None
        self._rows = None

    @classmethod
    def _create_new(cls, nixparent, h5parent,
                    name, type_, shape, col_dtype, compression):
        newentity = super(DataFrame, cls)._create_new(nixparent, h5parent,
                                                      name, type_)
        newentity._h5group.create_dataset("data", (shape, ), col_dtype)
        return newentity

    def append_column(self, column, name, datatype=None):
        """
        Append a new column to the DataFrame
        In case of string, it will be better to set explicitly the datatype.

        :param column: The new column
        :type column: array-like data
        :param name: The name of new column
        :type name: str
        :param datatype: The DataType of new column
        :type datatype: DataType
        """
        if len(column) < len(self):
            raise ValueError("Not enough entries for column in this dataframe")
        elif len(column) > len(self):
            raise ValueError("Too much entries for column in this dataframe")
        if datatype is None:
            datatype = DataType.get_dtype(column[0])
        if isclass(datatype) and any(issubclass(datatype, st)
                                     for st in string_types):
            datatype = util.vlen_str_dtype
        dt_arr = [(n, dty) for n, dty in zip(self.column_names, self.dtype)]
        dt_arr.append((name, datatype))
        dt = np.dtype(dt_arr)
        column = np.array(column, dtype=datatype)
        new_da = []
        for i, rows in enumerate(self._h5group.group['data'][:]):
            li = list(rows)
            li.append(column[i])
            tu = tuple(li)
            new_da.append(tu)
        farr = np.ascontiguousarray(new_da, dtype=dt)
        del self._h5group.group['data']
        self._h5group.group['data'] = farr
        self._h5group.create_dataset("data", (self.shape[0],), dt)
        self.write_direct(farr)

    def append_rows(self, data):
        """
        Append a new row to the DataFrame. The data supplied must be iterable.

        :param data: The new row
        :type data: array-like data
        """
        li_data = []
        for d in data:
            d = tuple(d)
            li_data.append(d)
        pro_data = np.array(li_data, dtype=self.data_type)
        self.append(pro_data, axis=0)

    def write_column(self, column, index=None, name=None):
        """
        Overwrite an existing column.
        Either index or name of the column should be provided.

        :param column: The new column
        :type column: array-like data
        :param index: The index of the column that is written to
        :type index: string
        :param name: The name of the column that is written to
        :type name: str
        """
        if len(column) != self.shape[0]:
            raise ValueError('If there are missing data, please fill in None')
        if not index and not name:
            raise ValueError("Either index or name must not be None")
        if name is None:
            name = self._find_name_by_idx(index)
        column = np.array(column)
        for i, rows in enumerate(self._h5group.group['data'][:]):
            cell = column[i]
            rows[name] = cell
            self.write_rows(rows=[rows], index=[i])

    def read_columns(self, index=None, name=None, sl=None,
                     group_by_cols=False):
        """
        Read one or multiple (part of) column(s) in the DataFrame

        :param index: Index of column(s) to be returned
        :type index: list of int
        :param name: Name of column(s) to be returned
        :type name: list of str
        :param sl: The part of each column to be returned
        :type sl: slice
        :param group_by_cols: True for group return values by columns,
                              False for group by rows.
                              Only applicable for reading multiple columns
        :type group_by_cols: bool
        """
        if index is None and name is None:
            raise ValueError("Either index or name must not be None")
        if name is None:
            name = []
            for ci in index:
                name.append(self.column_names[ci])
        if sl is None:
            slic = np.s_[:]
        else:
            slic = np.s_[sl]
        if len(name) == 1:
            get_col = self._read_data(sl=slic)[name]
            get_col = [i[0] for i in get_col]
            return np.array(get_col)
        if group_by_cols:
            gcol = []
            for n in name:
                get_col = self._read_data(sl=slic)[n]
                get_col = [i for i in get_col]
                gcol.append(get_col)
            return np.array(gcol)
        else:
            get_col = self._read_data(sl=slic)[name]
            return get_col

    def write_rows(self, rows, index):
        """
        Overwrite one or multiple existing row(s)

        :param rows: The new rows(s) and their data
        :type rows: (nested) array-like data
        :param index: Index of rows(s) to be overwritten
        :type index: list of int
        """
        if not isinstance(rows[0], (Iterable, np.void)):
            if len(index) != 1:
                raise TypeError("Rows should be in nested form")
            else:
                rows = [rows]
        if len(rows) != len(index):
            raise IndexError(
                "Number of rows ({}) does not match "
                "length of indexes ({})".format(len(rows), len(index))
            )
        x, = self.shape
        if max(index) > (x - 1):
            raise OutOfBounds(
                "Row index exceeds the existing number of rows"
            )
        if len(index) == 1:
            rows = tuple(rows[0])
            self._write_data(rows, sl=index)
        else:
            cr_list = []
            for i, cr in enumerate(rows):
                cr_list.append(tuple(cr))
            self._write_data(cr_list, sl=index)

    def read_rows(self, index):
        """
        Read one or multiple row(s) in the DataFrame

        :param index: Index of row(s) to be returned
        :type index: list of int
        """
        if isinstance(index, Iterable):
            index = list(index)
        get_row = self._read_data(sl=(index,))
        return get_row

    def write_cell(self, cell, position=None, col_name=None, row_idx=None):
        """
        Overwrite a cell in the DataFrame

        :param cell: The new cell
        :type cell: same type as the specified column
        :param position: Position of the targeted cell
        :type position: tuple or list or array with length 2
        :param col_name: The column name in which the targeted cell belongs to
        :type col_name: str
        :param row_idx: A length 1 list that specify on
                        which row the targeted cell is located
        :type row_idx: list of int
         """
        if position is not None:
            if len(position) != 2:
                raise ValueError("position is invalid: "
                                 "need row and column index")
            x, y = position
            targeted_row = self.read_rows(x)
            targeted_row[y] = cell
            self._write_data(targeted_row, sl=x)
        else:
            if col_name is None or row_idx is None:
                raise ValueError("Column and rows identifier must be given")
            targeted_row = self.read_rows(row_idx)
            targeted_row[col_name] = cell
            self._write_data(targeted_row, sl=row_idx)

    def read_cell(self, position=None, col_name=None, row_idx=None):
        """
        Read a cell in the DataFrame

        :param position: Position of the targeted cell
        :type position: tuple or list or array with length 2
        :param col_name: The column name in which the targeted cell belongs to
        :type col_name: str
        :param row_idx: A length 1 list that specify on
                         which row the targeted cell is located
        :type row_idx: list of int
        """
        if position is not None:
            if len(position) != 2:
                raise ValueError('Not a position')
            x, y = position
            return self[x][y]
        else:
            if col_name is None or row_idx is None:
                raise ValueError("Column and rows identifier must be given")
            cell = self[row_idx][col_name]
            cell = cell[0]
            return cell

    def print_table(self, row_sl=None, col_sl=None):
        """
        Print the whole DataFrame as a table
        :param row_sl: Rows to be printed; None for printing all rows
        :type row_sl: slice or iterable of int
        :param col_sl: Columns to be printed; None for printing all columns
        :type col_sl: slice or iterable of int
        """
        if row_sl is None:
            row_sl = np.s_[:]
        if col_sl is None:
            col_sl = np.s_[:]
        cl = np.array(self.column_names)[col_sl]
        row_form = "{:^10}" * (len(cl) + 1)
        print(row_form.format("column:", *cl))
        if self.units is not None:
            print(row_form.format(" unit:", *self.units[col_sl]))
        if not isinstance(row_sl, Iterable):
            ridx = list(range(len(self)))[row_sl]
        else:
            ridx = row_sl
        for i, row in enumerate(self._read_data(sl=row_sl)[list(cl)]):
            print(row_form.format("  [{}]:".format(ridx[i]), *row))

    def _find_idx_by_name(self, name):
        for i, n in enumerate(self.column_names):
            if n == name:
                return i
        return None

    def _find_name_by_idx(self, idx):
        if self.column_names[idx]:
            return self.column_names[idx]
        return None

    def row_count(self):
        """
        Return the total number of rows
        """
        count = len(self)
        return count

    def write_to_csv(self, filename, mode='w'):
        """
        Export the whole DataFrame to a CSV file

        :param filename: The resulted/ targeted CSV file to write to/ create
        :type filename: str
        :param mode: The column name in which the targeted cell belongs to
        :type mode: str
        """
        with open(filename, mode, newline='') as csvfile:
            dw = csv.DictWriter(csvfile, fieldnames=self.column_names)
            dw.writeheader()
            di = dict()
            for n in self.column_names:
                n = str(n)
                di[n] = list(self[n])
            complete_di_list = []
            sample_len = len(self[n])
            for i in range(sample_len):
                single_sample_di = dict()
                for na in self.column_names:
                    single_sample_di[na] = di[na][i]
                complete_di_list.append(single_sample_di)
            dw.writerows(complete_di_list)
            csvfile.close()

    @property
    def units(self):
        """
        The unit of the values stored in the DataFrame.
        This is a read-write property and can be set to None.

        :type: array of str
        """
        return self._h5group.get_attr("units")

    @units.setter
    def units(self, u):
        for i in u:
            if i is not None:
                i = util.units.sanitizer(i)
                util.check_attr_type(i, str)
        u = np.array(u, dtype=util.vlen_str_dtype)
        self._h5group.set_attr("units", u)
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    @property
    def columns(self):
        """
        The dtype is the list of names and data types
        of all columns in the DatFrame.
        This is a read only property.

        :type: list of tuples
        """
        if np.any(self.units):
            cols = [(n, dt, u) for n, dt, u in
                    zip(self.column_names, self.dtype, self.units)]
        else:
            cols = [(n, dt, None) for n, dt in
                    zip(self.column_names, self.dtype)]
        return cols

    @property
    def column_names(self):
        """
        The dtype is the list of names of all columns in the DatFrame.
        This is a read only property.

        :type: list of str
        """
        dt = self._h5group.group["data"].dtype
        return dt.names

    @property
    def dtype(self):
        """
        The dtype is the list of DataTypes of all columns in the DatFrame.
        This is a read only property.

        :type: list of DataType
        """
        dt = self._h5group.group["data"].dtype
        key = self.column_names
        di = OrderedDict()
        for k in key:
            di[k] = dt.fields[k]
        raw_dt = di.values()
        raw_dt = list(raw_dt)
        raw_dt_list = [ele[0] for ele in raw_dt]
        return tuple(raw_dt_list)

    @property
    def df_shape(self):
        """
        The df_shape is the shape of the DataFrame
        in (number of rows, number of columns) format.
        This is a read only property.

        :type: tuple
        """
        x = len(self._h5group.group["data"])
        y = len(self.column_names)
        df_shape = (x, y)
        df_shape = tuple(df_shape)
        self._h5group.set_attr("df_shape", df_shape)
        return self._h5group.get_attr("df_shape")

    @property
    def metadata(self):
        """
        Associated metadata of the entity. Sections attached to the entity via
        this attribute can provide additional annotations. This is an optional
        read-write property, and can be None if no metadata is available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("{} is not of type Section".format(sect))
        self._h5group.create_link(sect, "metadata")

    @metadata.deleter
    def metadata(self):
        if "metadata" in self._h5group:
            self._h5group.delete("metadata")



from .entity import Entity
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from . import util


class Container(object):
    """
    Container acts as an interface to container groups in the backend. In the
    case of HDF5, this is a group that is used as a container for other groups.

    Examples of Containers:
        Block.data_arrays
        Block.tags
        Block.multi_tags
        Source.sources

    :param name: Name of the container
    :param parent: Parent NIX object where this container will be created
    :param itemclass: The class of the objects this container holds (for
    checking and instantiations)
    """

    def __init__(self, name, parent, itemclass):
        self._backend = parent._h5group.open_group(name)
        self._itemclass = itemclass
        self._parent = parent
        self._name = name

    def _inst_item(self, item):
        return self._itemclass(self._parent, item)

    def __len__(self):
        return len(self._backend)

    def __getitem__(self, item):
        if isinstance(item, int):
            if item < 0:
                item = len(self) + item
            if item < 0 or item >= len(self):
                raise IndexError("Index out of bounds: {}".format(item))
            item = self._backend.get_by_pos(item)
        else:
            item = self._backend.get_by_id_or_name(item)
        return self._inst_item(item)

    def __delitem__(self, item):
        if not isinstance(item, Entity):
            item = self[item]

        if not isinstance(item, self._itemclass):
            raise TypeError(
                "Wrong item type: {} required or the name or ID of one".format(
                    self._itemclass.__name__)
            )

        root = self._backend.h5root
        if not root:
            root = self._parent._h5group
        root.delete_all([item.id])

    def __iter__(self):
        for group in self._backend:
            yield self._inst_item(group)

    def __contains__(self, item):
        if hasattr(item, "id"):
            if isinstance(item, self._itemclass):
                return item.name in self._backend
            # looks like a NIX object, but wrong type
            raise TypeError(
                "Wrong item type: {} required or the name or ID of one".format(
                    self._itemclass.__name__)
            )
        if util.is_uuid(item):
            try:
                self._backend.get_by_id(item)
                return True
            except KeyError:
                return False
        else:
            return item in self._backend

    def __str__(self):
        return "[{}]".format(
            ", ".join(str(item) for item in self)
        )

    def __repr__(self):
        return str(self)

    @staticmethod
    def _item_key(item):
        return item.name

    def items(self):
        for group in self._backend:
            item = self._inst_item(group)
            yield item.id, item


class SectionContainer(Container):
    """
    SectionContainer extends Container with a new __delitem__ method.
    When a Section is deleted, all child sections need to be deleted
    individually to make sure all their references are removed.
    """
    def __delitem__(self, item):
        if not isinstance(item, Entity):
            item = self[item]

        if not isinstance(item, self._itemclass):
            raise TypeError(
                "Wrong item type: {} required or the name or ID of one".format(
                    self._itemclass.__name__)
            )

        # collect all IDs under item and send them for deletion, starting from
        # the root block
        secids = [s.id for s in item.find_sections()]

        root = self._backend.file
        root.delete_all(secids)


class SourceContainer(Container):
    """
    SourceContainer extends Container with a new __delitem__ method.
    When a Source is deleted, all child sources need to be deleted individually
    to make sure all their references are removed.
    """
    def __delitem__(self, item):
        if not isinstance(item, Entity):
            item = self[item]

        if not isinstance(item, self._itemclass):
            raise TypeError(
                "Wrong item type: {} required or the name or ID of one".format(
                    self._itemclass.__name__)
            )

        # collect all IDs under item and send them for deletion, starting from
        # the root block
        srcids = [s.id for s in item.find_sources()]
        srcids.append(item.id)

        root = self._backend.h5root
        root.delete_all(srcids)


class LinkContainer(Container):
    """
    A LinkContainer acts as an interface to container groups in the backend
    that hold links to objects already contained in a Container.

    Objects are added to a LinkContainer using the 'append' method, as
    opposed to Containers which get populated by the parent object 'create'
    methods.

    An important difference between a LinkContainer and a Container is that
    links to objects are indexed by their 'id' whereas objects in Containers
    are indexed by 'name'.

    Examples of LinkContainers:
        Group.data_arrays
        Group.tags
        Group.multi_tags

    :param name: Name of the container
    :param parent: Parent H5Group where this container will be created
    :param itemclass: The class of the objects this container holds (for
    checking and instantiations)
    :param itemstore: The location (Container) where the original objects
    are stored and linked to.
    """

    def __init__(self, name, parent, itemclass, itemstore):
        super(LinkContainer, self).__init__(name, parent, itemclass)
        self._itemstore = itemstore

    def __delitem__(self, item):
        if not isinstance(item, Entity):
            item = self[item]

        if not isinstance(item, self._itemclass):
            raise TypeError(
                "Wrong item type: {} required or the name or ID of one".format(
                    self._itemclass.__name__)
            )

        self._backend.delete(item.id)

    def append(self, item):
        if util.is_uuid(item):
            item = self._inst_item(self._backend.get_by_id(item))

        if not hasattr(item, "id"):
            raise TypeError("NIX entity or id string required for append")

        if item not in self._itemstore:
            raise RuntimeError("This item cannot be appended here.")

        self._backend.create_link(item, item.id)

    def extend(self, items):
        if not isinstance(items, Iterable):
            raise TypeError("{} object is not iterable".format(type(items)))
        for item in items:
            self.append(item)

    def __getitem__(self, identifier):
        if isinstance(identifier, int):
            return super(LinkContainer, self).__getitem__(identifier)
        else:
            if util.is_uuid(identifier):
                # For LinkContainer, name is id
                item = self._backend.get_by_name(identifier)
                return self._inst_item(item)
            else:
                for grp in self._backend:
                    if identifier == grp.get_attr("name"):
                        return self._inst_item(grp)

                raise KeyError("Item not found '{}'".format(identifier))

    def __contains__(self, item):
        # need to redefine because of id indexing/linking
        if hasattr(item, "id"):
            if isinstance(item, self._itemclass):
                return item.id in self._backend
            # looks like a NIX object, but wrong type
            raise TypeError(
                "Wrong item type: {} required or the name or ID of one".format(
                    self._itemclass.__name__)
            )

        if util.is_uuid(item):
            return item in self._backend

        # assume it's a name and scan through LinkContainer
        for grp in self._backend:
            if item == grp.get_attr("name"):
                return True
        return False

    def _inst_item(self, item):
        return self._itemclass(self._itemstore._parent, item)

    @staticmethod
    def _item_key(item):
        return item.id



# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from sys import maxsize as maxint

from .import exceptions
from .entity import Entity
from .container import SourceContainer
from . import util
from .util import find as finders
from .section import Section


class Source(Entity):

    def __init__(self, nixparent, h5group):
        super(Source, self).__init__(nixparent, h5group)
        self._sources = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(Source, cls)._create_new(nixparent, h5parent,
                                                   name, type_)
        return newentity

    # Source
    def create_source(self, name, type_):
        """
        Create a new source as a child of the current Source.

        :param name: The name of the source to create.
        :type name: str
        :param type_: The type of the source.
        :type type_: str

        :returns: The newly created source.
        :rtype: Source
        """
        util.check_entity_name_and_type(name, type_)
        sources = self._h5group.open_group("sources", True)
        if name in sources:
            raise exceptions.DuplicateName("create_source")
        src = Source._create_new(self, sources, name, type_)
        return src

    @property
    def referring_objects(self):
        objs = []
        objs.extend(self.referring_data_arrays)
        objs.extend(self.referring_tags)
        objs.extend(self.referring_multi_tags)
        return objs

    @property
    def referring_data_arrays(self):
        return [da for da in self._parent.data_arrays if self in da.sources]

    @property
    def referring_tags(self):
        return [tg for tg in self._parent.tags if self in tg.sources]

    @property
    def referring_multi_tags(self):
        return [mt for mt in self._parent.multi_tags if self in mt.sources]

    def find_sources(self, filtr=lambda _: True, limit=None):
        """
        Get all child sources of this source recursively.

        This method traverses the tree of all sources. The traversal
        is accomplished via breadth first and can be limited in depth. On each
        node or source a filter is applied. If the filter returns true the
        respective source will be added to the result list.
        By default a filter is used that accepts all sources.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sources.
        :rtype: list of Source
        """
        if limit is None:
            limit = maxint
        return finders._find_sources(self, filtr, limit)

    @property
    def sources(self):
        """
        A property containing child sources of a Source. Sources can be
        obtained via their name, index, id. Sources can be deleted from the
        list. Adding sources is done using the Blocks create_source method.
        This is a read only attribute.
        """
        if self._sources is None:
            self._sources = SourceContainer("sources", self, Source)
        return self._sources

    # metadata
    @property
    def metadata(self):
        """

        Associated metadata of the entity. Sections attached to the entity via
        this attribute can provide additional annotations. This is an optional
        read-write property, and can be None if no metadata is available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("{} is not of type Section".format(sect))
        self._h5group.create_link(sect, "metadata")

    @metadata.deleter
    def metadata(self):
        if "metadata" in self._h5group:
            self._h5group.delete("metadata")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how multiple regularly sampled data that belong
 together can be stored in a single DataArray in nix-files. See
 https://github.com/G-node/nix/wiki for more information.

"""

import nixio as nix
import numpy as np
import matplotlib.pylab as plt


def create_data(duration=1, freq=10, stepsize=0.01):
    x = np.arange(0, duration*2*np.pi, stepsize)
    sine = np.sin(freq*x)
    cosine = np.cos(freq*x)
    return x, sine, cosine


def plot_data(data_array):
    set_dim = data_array.dimensions[0]
    x_axis = data_array.dimensions[1]
    x = np.arange(0, data_array.data.shape[1])
    x = x * x_axis.sampling_interval + x_axis.offset
    y = np.zeros(data_array.data.shape)
    data_array.data.read_direct(y)
    for i, label in enumerate(set_dim.labels):
        plt.plot(x, y[i, :], label=label)

    plt.xlabel(x_axis.label + " [" + x_axis.unit + "]")
    plt.ylabel(data_array.label + " [" + data_array.unit + "]")
    plt.title(data_array.name)
    plt.xlim(0, np.max(x))
    plt.ylim((1.1 * np.min(y), 1.1 * np.max(y)))
    plt.legend()
    plt.savefig('multiple_time_series.png')
    plt.show()


if __name__ == "__main__":
    # fake some data
    duration = 2
    frequency = 2
    stepsize = 0.02
    x, sine, cosine = create_data(duration, frequency, stepsize)

    # create a new file overwriting any existing content
    file_name = 'multiple_regular_data_example.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")
    # create a 'DataArray' to take the data, add some information about the signal
    y = np.vstack((sine, cosine))
    data = block.create_data_array("waveforms", "nix.regular_sampled.multiple_series", data=y)
    data.unit = "mV"
    data.label = "voltage"
    # descriptor for first dimension is a set
    set_dim = data.append_set_dimension()
    set_dim.labels = ['sin', 'cos']
    # add a descriptor for the xaxis
    dim = data.append_sampled_dimension(stepsize)
    dim.unit = "s"
    dim.label = "time"
    dim.offset = 0.0 # optional

    # let's plot the data from the stored information
    plot_data(data)
    file.close()



# -*- coding: utf-8 -*-
# Copyright © 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
import warnings
from numbers import Number
from enum import Enum

from .data_view import DataView
from .data_set import DataSet
from .entity import Entity
from .source_link_container import SourceLinkContainer
from .datatype import DataType
from .dimensions import (Dimension, SampledDimension, RangeDimension,
                         SetDimension, DimensionType, DimensionContainer)
from . import util
from .compression import Compression

from .exceptions import InvalidUnit, IncompatibleDimensions
from .section import Section


class DataSliceMode(Enum):
    Index = 1
    Data = 2


class DataArray(Entity, DataSet):

    def __init__(self, nixparent, h5group):
        super(DataArray, self).__init__(nixparent, h5group)
        self._sources = None
        self._dimensions = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, data_type, shape,
                    compression):
        newentity = super(DataArray, cls)._create_new(nixparent, h5parent,
                                                      name, type_)
        datacompr = False
        if compression == Compression.DeflateNormal:
            datacompr = True
        newentity._h5group.create_dataset("data", shape, data_type, datacompr)
        return newentity

    def _read_data(self, sl=None):
        coeff = self.polynom_coefficients
        origin = self.expansion_origin
        sup = super(DataArray, self)
        if len(coeff) or origin:
            if not origin:
                origin = 0.0

            # when there are coefficients, convert the dtype of the returned
            # data array to double
            data = sup._read_data(sl).astype(DataType.Double)
            util.apply_polynomial(coeff, origin, data)
        else:
            data = sup._read_data(sl)
        return data

    @property
    def sources(self):
        """
        A property containing all Sources referenced by the DataArray. Sources
        can be obtained by index or their id. Sources can be removed from the
        list, but removing a referenced Source will not remove it from the
        file. New Sources can be added using the append method of the list.
        This is a read only attribute.
        """
        if self._sources is None:
            self._sources = SourceLinkContainer(self)
        return self._sources

    def append_set_dimension(self, labels=None):
        """
        Append a new SetDimension to the list of existing dimension
        descriptors.

        :returns: The newly created SetDimension.
        :rtype: SetDimension
        """
        dimgroup = self._h5group.open_group("dimensions")
        index = len(dimgroup) + 1
        setdim = SetDimension._create_new(dimgroup, index)
        if labels:
            setdim.labels = labels
        if self._parent._parent.time_auto_update:
            self.force_updated_at()
        return setdim

    def append_sampled_dimension(self, sampling_interval, label=None,
                                 unit=None, offset=None):
        """
        Append a new SampledDimension to the list of existing dimension
        descriptors.

        :param sampling_interval: The sampling interval of the SetDimension
                                  to create.
        :type sampling_interval: float

        :returns: The newly created SampledDimension.
        :rtype: SampledDimension
        """
        dimgroup = self._h5group.open_group("dimensions")
        index = len(dimgroup) + 1
        smpldim = SampledDimension._create_new(dimgroup, index,
                                               sampling_interval)
        if label:
            smpldim.label = label
        if unit:
            smpldim.unit = unit
        if offset:
            smpldim.offset = offset
        if self._parent._parent.time_auto_update:
            self.force_updated_at()
        return smpldim

    def append_range_dimension(self, ticks, label=None, unit=None):
        """
        Append a new RangeDimension to the list of existing dimension
        descriptors.

        :param ticks: The ticks of the RangeDimension to create.
        :type ticks: list of float

        :returns: The newly created RangeDimension.
        :rtype: RangeDimension
        """
        dimgroup = self._h5group.open_group("dimensions")
        index = len(dimgroup) + 1
        rdim = RangeDimension._create_new(dimgroup, index, ticks)
        if label:
            rdim.label = label
            rdim.unit = unit
        if self._parent._parent.time_auto_update:
            self.force_updated_at()
        return rdim

    def append_alias_range_dimension(self):
        """
        Append a new RangeDimension that uses the data stored in this
        DataArray as ticks. This works only(!) if the DataArray is 1-D and
        the stored data is numeric. A ValueError will be raised otherwise.

        :returns: The created dimension descriptor.
        :rtype: RangeDimension
        """
        if (len(self.data_extent) > 1 or
                not DataType.is_numeric_dtype(self.dtype)):
            raise ValueError("AliasRangeDimensions only allowed for 1D "
                             "numeric DataArrays.")
        if self._dimension_count() > 0:
            raise ValueError("Cannot append additional alias dimension. "
                             "There must only be one!")
        dimgroup = self._h5group.open_group("dimensions")
        # check if existing unit is SI
        if self.unit:
            u = self.unit
            if not (util.units.is_si(u) or util.units.is_compound(u)):
                raise InvalidUnit(
                    "AliasRangeDimensions are only allowed when SI or "
                    "composites of SI units are used. "
                    "Current SI unit is {}".format(u),
                    "DataArray.append_alias_range_dimension"
                )
        return RangeDimension._create_new_alias(dimgroup, 1, self)

    def delete_dimensions(self):
        """
        Delete all the dimension descriptors for this DataArray.
        """
        dimgroup = self._h5group.open_group("dimensions")
        ndims = len(dimgroup)
        for idx in range(ndims):
            del dimgroup[str(idx+1)]
        return True

    def _dimension_count(self):
        return len(self._h5group.open_group("dimensions"))

    def _get_dimension_by_pos(self, index):
        h5dim = self._h5group.open_group("dimensions").open_group(str(index))
        dimtype = h5dim.get_attr("dimension_type")
        if DimensionType(dimtype) == DimensionType.Sample:
            return SampledDimension(h5dim, index)
        elif DimensionType(dimtype) == DimensionType.Range:
            return RangeDimension(h5dim, index)
        elif DimensionType(dimtype) == DimensionType.Set:
            return SetDimension(h5dim, index)
        else:
            raise TypeError("Invalid Dimension object in file.")

    @property
    def dtype(self):
        """
        The data type of the data stored in the DataArray.
        This is a read only property.

        :return: DataType
        """
        return self._h5group.group["data"].dtype

    @property
    def polynom_coefficients(self):
        """
        The polynomial coefficients for the calibration. By default this is
        set to a {0.0, 1.0} for a linear calibration with zero offset.
        This is a read-write property and can be set to None

        :type: list of float
        """
        return tuple(self._h5group.get_data("polynom_coefficients"))

    @polynom_coefficients.setter
    def polynom_coefficients(self, coeff):
        if not coeff:
            if self._h5group.has_data("polynom_coefficients"):
                del self._h5group["polynom_coefficients"]
        else:
            dtype = DataType.Double
            self._h5group.write_data("polynom_coefficients", coeff, dtype)
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    @property
    def expansion_origin(self):
        """
        The expansion origin of the calibration polynomial.
        This is a read-write property and can be set to None.
        The default value is 0.

        :type: float
        """
        return self._h5group.get_attr("expansion_origin")

    @expansion_origin.setter
    def expansion_origin(self, eo):
        util.check_attr_type(eo, Number)
        self._h5group.set_attr("expansion_origin", eo)
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    @property
    def label(self):
        """
        The label of the DataArray. The label corresponds to the label of the
        x-axis of a plot. This is a read-write property and can be set to
        None.

        :type: str
        """
        return self._h5group.get_attr("label")

    @label.setter
    def label(self, l):
        util.check_attr_type(l, str)
        self._h5group.set_attr("label", l)
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    @property
    def unit(self):
        """
        The unit of the values stored in the DataArray. This is a read-write
        property and can be set to None.

        :type: str
        """
        return self._h5group.get_attr("unit")

    @unit.setter
    def unit(self, u):
        if u:
            u = util.units.sanitizer(u)
        if u == "":
            u = None
        util.check_attr_type(u, str)
        if (self._dimension_count() == 1 and
            self.dimensions[0].dimension_type == DimensionType.Range and
                self.dimensions[0].is_alias and u is not None):
            if not (util.units.is_si(u) or util.units.is_compound(u)):
                raise InvalidUnit(
                    "[{}]: Non-SI units are not allowed if the DataArray "
                    "has an AliasRangeDimension.".format(u),
                    "DataArray.unit"
                )
        self._h5group.set_attr("unit", u)
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    def get_slice(self, positions, extents=None, mode=DataSliceMode.Index):
        datadim = len(self.shape)
        if not len(positions) == datadim:
            raise IncompatibleDimensions(
                "Number of positions given ({}) does not match "
                "number of data dimensions ({})".format(
                    len(positions), datadim
                ),
                "DataArray.get_slice"
            )
        if extents and not len(extents) == datadim:
            raise IncompatibleDimensions(
                "Number of extents given ({}) does not match "
                "number of data dimensions ({})".format(
                    len(extents), datadim
                ),
                "DataArray.get_slice"
            )
        if mode == DataSliceMode.Index:
            sl = tuple(slice(p, p+e) for p, e in zip(positions, extents))
            return DataView(self, sl)
        elif mode == DataSliceMode.Data:
            return self._get_slice_bydim(positions, extents)
        else:
            raise ValueError("Invalid slice mode specified. "
                             "Supported modes are DataSliceMode.Index and "
                             "DataSliceMode.Data")

    def _get_slice_bydim(self, positions, extents):
        dpos, dext = [], []
        for dim, pos, ext in zip(self.dimensions, positions, extents):
            if dim.dimension_type in (DimensionType.Sample,
                                      DimensionType.Range):
                dpos.append(dim.index_of(pos))
                dext.append(dim.index_of(pos+ext)-dpos[-1])
            elif dim.dimension_type == DimensionType.Set:
                dpos.append(int(pos))
                dext.append(int(ext))
        sl = tuple(slice(p, p+e) for p, e in zip(dpos, dext))
        return DataView(self, sl)

    @property
    def data(self):
        """
        DEPRECATED DO NOT USE ANYMORE! Returns self

        :type: :class:`~nixio.data_array.DataArray`
        """
        warnings.warn("Call to deprecated property DataArray.data",
                      category=DeprecationWarning)
        return self

    @property
    def dimensions(self):
        """
        A property containing all dimensions of a DataArray. Dimensions can be
        obtained via their index. Adding dimensions is done using the
        respective append methods for dimension descriptors.
        This is a read only attribute.

        :type: Container of dimension descriptors.
        """
        if self._dimensions is None:
            self._dimensions = DimensionContainer("dimensions", self,
                                                  Dimension)
        return self._dimensions

    # metadata
    @property
    def metadata(self):
        """

        Associated metadata of the entity. Sections attached to the entity via
        this attribute can provide additional annotations. This is an optional
        read-write property, and can be None if no metadata is available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("{} is not of type Section".format(sect))
        self._h5group.create_link(sect, "metadata")

    @metadata.deleter
    def metadata(self):
        if "metadata" in self._h5group:
            self._h5group.delete("metadata")



# Copyright (c) 2017, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from .source import Source
from .container import LinkContainer
from . import util


class SourceLinkContainer(LinkContainer):

    # TODO: Sources returned from this container have an incorrect _parent
    # This is the same issue that we have with Sections. It should probably be
    # solved the same way

    # NOTE: Source requires a different link container since a valid Source
    # may exist anywhere beneath the root Block's child sources. The check for
    # the Source's validity needs to be done with the find_sources() function
    # down the entire tree, as opposed to the default LinkContainer.append()
    # function which needs to only check the root Block's container.

    def __init__(self, parent):
        """
        LinkContainer for Sources.
        """
        super(SourceLinkContainer, self).__init__("sources", parent, Source,
                                                  parent._parent.sources)

    def append(self, item):
        if util.is_uuid(item):
            item = self._inst_item(self._backend.get_by_id(item))

        if not hasattr(item, "id"):
            raise TypeError("NIX entity or id string required for append")

        if not self._itemstore._parent.find_sources(filtr=lambda x:
                                                    x.id == item.id):
            raise RuntimeError("This item cannot be appended here.")

        self._backend.create_link(item, item.id)



# -*- coding: utf-8 -*-
# Copyright © 2015, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from .entity import Entity
from .data_array import DataArray
from .data_frame import DataFrame
from .tag import Tag
from .multi_tag import MultiTag
from .container import LinkContainer
from .source_link_container import SourceLinkContainer
from .section import Section


class Group(Entity):

    def __init__(self, nixparent, h5group):
        super(Group, self).__init__(nixparent, h5group)
        self._data_arrays = None
        self._data_frames = None
        self._tags = None
        self._multi_tags = None
        self._sources = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(Group, cls)._create_new(nixparent, h5parent,
                                                  name, type_)
        return newentity

    @property
    def data_arrays(self):
        """
        A property containing all data arrays referenced by the group.
        Referenced data arrays can be obtained by index or their id. References
        can be removed from the list, removing a referenced DataArray will not
        remove it from the file. New references can be added using the append
        method of the list.
        This is a read only attribute.
        """
        if self._data_arrays is None:
            self._data_arrays = LinkContainer("data_arrays", self, DataArray,
                                              self._parent.data_arrays)
        return self._data_arrays

    @property
    def data_frames(self):
        """
        A property containing all data frames referenced by the group.
        Referenced data frames can be obtained by index or their id. References
        can be removed from the list, removing a referenced DataFrame will not
        remove it from the file. New references can be added using the append
        method of the list.
        This is a read only attribute.
        """
        if self._data_frames is None:
            self._data_frames = LinkContainer("data_frames", self, DataFrame,
                                              self._parent.data_frames)
        return self._data_frames

    @property
    def tags(self):
        """
        A property containing all tags referenced by the group. Tags can be
        obtained by index or their id. Tags can be removed from the list,
        removing a referenced Tag will not remove it from the file.
        New Tags can be added using the append method of the list.
        This is a read only attribute.
        """
        if self._tags is None:
            self._tags = LinkContainer("tags", self, Tag, self._parent.tags)
        return self._tags

    @property
    def multi_tags(self):
        """
        A property containing all MultiTags referenced by the group. MultiTags
        can be obtained by index or their id. MultiTags can be removed from the
        list, removing a referenced MultiTag will not remove it from the file.
        New MultiTags can be added using the append method of the list.
        This is a read only attribute.
        """
        if self._multi_tags is None:
            self._multi_tags = LinkContainer("multi_tags", self, MultiTag,
                                             self._parent.multi_tags)
        return self._multi_tags

    @property
    def sources(self):
        """
        A property containing all Sources referenced by the Group. Sources
        can be obtained by index or their id. Sources can be removed from the
        list, but removing a referenced Source will not remove it from the
        file. New Sources can be added using the append method of the list.
        This is a read only attribute.
        """
        if self._sources is None:
            self._sources = SourceLinkContainer(self)
        return self._sources

    # metadata
    @property
    def metadata(self):
        """

        Associated metadata of the entity. Sections attached to the entity via
        this attribute can provide additional annotations. This is an optional
        read-write property, and can be None if no metadata is available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("{} is not of type Section".format(sect))
        self._h5group.create_link(sect, "metadata")

    @metadata.deleter
    def metadata(self):
        if "metadata" in self._h5group:
            self._h5group.delete("metadata")



# -*- coding: utf-8 -*-
# Copyright © 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
import numpy as np
from six import ensure_str
from . import util


class DataSet(object):
    """
    Data IO object for DataArray.
    """

    def __array__(self):
        return self._read_data()[:]

    def __getitem__(self, index):
        return self._read_data(index)

    def __setitem__(self, index, value):
        self._write_data(value, index)

    def __len__(self):
        return self.len()

    def __iter__(self):
        for idx in range(self.len()):
            yield self[idx]

    def len(self):
        """
        Length of the first dimension. Equivalent to `DataSet.shape[0]`.

        :type: int or long
        """
        return self.shape[0]

    @property
    def shape(self):
        """
        :type: tuple of data array dimensions.
        """
        return self.data_extent

    @property
    def size(self):
        """
        Number of elements in the DataSet, i.e. the product of the
        elements in :attr:`~nixio.data_array.DataSet.shape`.

        :type: int
        """
        return np.prod(self.shape)

    @property
    def dtype(self):
        """
        :type: :class:`numpy.dtype` object holding type infromation about
               the data stored in the DataSet.
        """
        return np.dtype(self._get_dtype())

    def write_direct(self, data):
        """
        Directly write all of ``data`` to the
        :class:`~nixio.data_array.DataSet`.  The supplied data must be a
        :class:`numpy.ndarray` that matches the DataSet's shape and must have
        C-style contiguous memory layout (see :attr:`numpy.ndarray.flags` and
        :class:`~numpy.ndarray` for more information).

        :param data: The array which contents is being written
        :type data: :class:`numpy.ndarray`
        """
        self._write_data(data)

    def read_direct(self, data):
        """
        Directly read all data stored in the :class:`~nixio.data_array.DataSet`
        into ``data``. The supplied data must be a :class:`numpy.ndarray` that
        matches the DataSet's shape, must have C-style contiguous memory layout
        and must be writeable (see :attr:`numpy.ndarray.flags` and
        :class:`~numpy.ndarray` for more information).

        :param data: The array where data is being read into
        :type data: :class:`numpy.ndarray`
        """
        data[:] = self._read_data()

    def append(self, data, axis=0):
        """
        Append ``data`` to the DataSet along the ``axis`` specified.

        :param data: The data to append. Shape must agree except for the
                     specified axis
        :param axis: Along which axis to append the data to
        """
        data = np.ascontiguousarray(data)
        if len(self.shape) != len(data.shape):
            raise ValueError(
                "Data and DataArray must have the same dimensionality"
            )

        if any([s != ds for i, (s, ds) in
                enumerate(zip(self.shape, data.shape)) if i != axis]):
            raise ValueError("Shape of data and shape of DataArray must match "
                             "in all dimension but axis!")

        offset = tuple(0 if i != axis else x for i, x in enumerate(self.shape))
        count = data.shape
        enlarge = tuple(self.shape[i] + (0 if i != axis else x)
                        for i, x in enumerate(data.shape))
        self.data_extent = enlarge
        sl = tuple(slice(o, c+o) for o, c in zip(offset, count))
        self._write_data(data, sl)

    def _write_data(self, data, sl=None):
        dataset = self._h5group.get_dataset("data")
        dataset.write_data(data,  sl)

    def _read_data(self, sl=None):
        dataset = self._h5group.get_dataset("data")
        data = dataset.read_data(sl)
        if data.dtype == util.vlen_str_dtype:
            data = np.array([ensure_str(d) for d in data],
                            dtype=util.vlen_str_dtype)
        return data

    @property
    def data_extent(self):
        """
        The size of the data.

        :type: tuple of int
        """
        dataset = self._h5group.get_dataset("data")
        return dataset.shape

    @data_extent.setter
    def data_extent(self, extent):
        dataset = self._h5group.get_dataset("data")
        dataset.shape = extent

    @property
    def data_type(self):
        """
        The data type of the data stored in the DataArray. This is a read only
        property.

        :type: DataType
        """
        return self._get_dtype()

    def _get_dtype(self):
        dataset = self._h5group.get_dataset("data")
        return dataset.dtype



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import nixio as nix


file_name = 'file_create_example.h5'

# create a new file overwriting any existing content
nixfile = nix.File.open(file_name, nix.FileMode.Overwrite)
print(nixfile.format, nixfile.version, nixfile.created_at)

# close file
nixfile.close()

# re-open file for read-only access
nixfile = nix.File.open(file_name, nix.FileMode.ReadOnly)

# this command will fail putting out HDF5 Errors
try:
    nixfile.create_block("test block", "test")
except ValueError:
    print("Error caught: cannot create a new group in nix.FileMode.ReadOnly mode")


nixfile.close()

# re-open for read-write access
nixfile = nix.File.open(file_name, nix.FileMode.ReadWrite)

# the following command now works fine
nixfile.create_block("test block", "test")

nixfile.close()









# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

try:
    from sys import maxint
except ImportError:
    from sys import maxsize as maxint
import numpy as np
from inspect import isclass
from six import string_types
try:
    from collections.abc import OrderedDict
except ImportError:
    from collections import OrderedDict
import sys

from .util import find as finders
from .compression import Compression

from .entity import Entity
from .exceptions import exceptions
from .group import Group
from .data_array import DataArray
from .data_frame import DataFrame
from .multi_tag import MultiTag
from .tag import Tag
from .source import Source
from . import util
from .container import Container, SourceContainer
from .section import Section


class Block(Entity):

    def __init__(self, nixparent, h5group, compression=Compression.Auto):
        super(Block, self).__init__(nixparent, h5group)
        self._groups = None
        self._data_arrays = None
        self._tags = None
        self._multi_tags = None
        self._sources = None
        self._compr = compression
        self._data_frames = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, compression):
        newentity = super(Block, cls)._create_new(nixparent, h5parent,
                                                  name, type_)
        newentity._compr = compression
        return newentity

    # MultiTag
    def create_multi_tag(self, name="", type_="", positions=0,
                         copy_from=None, keep_copy_id=True):
        """
        Create/copy a new multi tag for this block.

        :param name: The name of the tag to create/copy.
        :type name: str
        :param type_: The type of tag.
        :type type_: str
        :param positions: A data array defining all positions of the tag.
        :type positions: DataArray
        :param copy_from: The MultiTag to be copied, None in normal mode
        :type copy_from: MultiTag
        :param keep_copy_id: Specify if the id should be copied in copy mode
        :type keep_copy_id: bool

        :returns: The newly created tag.
        :rtype: MultiTag
        """
        if copy_from:
            if not isinstance(copy_from, MultiTag):
                raise TypeError("Object to be copied is not a MultiTag")
            id = self._copy_objects(copy_from, "multi_tags",
                                    keep_copy_id, name)
            return self.multi_tags[id]

        util.check_entity_name_and_type(name, type_)
        util.check_entity_input(positions)
        if not isinstance(positions, DataArray):
            raise TypeError("DataArray expected for 'positions'")
        multi_tags = self._h5group.open_group("multi_tags")
        if name in multi_tags:
            raise exceptions.DuplicateName("create_multi_tag")
        mtag = MultiTag._create_new(self, multi_tags, name, type_, positions)
        return mtag

    # Tag
    def create_tag(self, name="", type_="", position=0,
                   copy_from=None, keep_copy_id=True):
        """
        Create/copy a new tag for this block.

        :param name: The name of the tag to create/copy.
        :type name: str
        :param type_: The type of tag.
        :type type_: str
        :param position: Coordinates of the start position
                         in units of the respective data dimension.
        :param copy_from: The Tag to be copied, None in normal mode
        :type copy_from: Tag
        :param keep_copy_id: Specify if the id should be copied in copy mode
        :type keep_copy_id: bool

        :returns: The newly created tag.
        :rtype: Tag
        """
        if copy_from:
            if not isinstance(copy_from, Tag):
                raise TypeError("Object to be copied is not a Tag")
            id = self._copy_objects(copy_from, "tags", keep_copy_id, name)
            return self.tags[id]

        util.check_entity_name_and_type(name, type_)
        tags = self._h5group.open_group("tags")
        if name in tags:
            raise exceptions.DuplicateName("create_tag")
        tag = Tag._create_new(self, tags, name, type_, position)
        return tag

    # Source
    def create_source(self, name, type_):
        """
        Create a new source on this block.

        :param name: The name of the source to create.
        :type name: str
        :param type_: The type of the source.
        :type type_: str

        :returns: The newly created source.
        :rtype: Source
        """
        util.check_entity_name_and_type(name, type_)
        sources = self._h5group.open_group("sources")
        if name in sources:
            raise exceptions.DuplicateName("create_source")
        src = Source._create_new(self, sources, name, type_)
        return src

    # Group
    def create_group(self, name, type_):
        """
        Create a new group on this block.

        :param name: The name of the group to create.
        :type name: str
        :param type_: The type of the group.
        :type type_: str

        :returns: The newly created group.
        :rtype: Group
        """
        util.check_entity_name_and_type(name, type_)
        groups = self._h5group.open_group("groups")
        if name in groups:
            raise exceptions.DuplicateName("open_group")
        grp = Group._create_new(self, groups, name, type_)
        return grp

    def create_data_array(self, name="", array_type="", dtype=None, shape=None,
                          data=None, compression=Compression.Auto,
                          copy_from=None, keep_copy_id=True):
        """
        Create/copy a new data array for this block. Either ``shape``
        or ``data`` must be given. If both are given their shape must agree.
        If ``dtype`` is not specified it will default to 64-bit floating
        points.

        :param name: The name of the data array to create/copy.
        :type name: str
        :param array_type: The type of the data array.
        :type array_type: str
        :param dtype: Which data-type to use for storage
        :type dtype:  :class:`numpy.dtype`
        :param shape: Layout (dimensionality and extent)
        :type shape: tuple of int or long
        :param data: Data to write after storage has been created
        :type data: array-like data
        :param compression: En-/disable dataset compression.
        :type compression: :class:`~nixio.Compression`
        :param copy_from: The DataArray to be copied, None in normal mode
        :type copy_from: DataArray
        :param keep_copy_id: Specify if the id should be copied in copy mode
        :type keep_copy_id: bool

        :returns: The newly created data array.
        :rtype: :class:`~nixio.DataArray`
        """

        if copy_from:
            if not isinstance(copy_from, DataArray):
                raise TypeError("Object to be copied is not a DataArray")
            id = self._copy_objects(copy_from, "data_arrays",
                                    keep_copy_id, name)
            return self.data_arrays[id]

        if data is None:
            if shape is None:
                raise ValueError("Either shape and or data must not be None")
            if dtype is None:
                dtype = 'f8'
        else:
            data = np.ascontiguousarray(data)
            if dtype is None:
                dtype = data.dtype
            if shape is not None:
                if shape != data.shape:
                    raise ValueError("Shape must equal data.shape")
            else:
                shape = data.shape
        util.check_entity_name_and_type(name, array_type)
        data_arrays = self._h5group.open_group("data_arrays")
        if name in data_arrays:
            raise exceptions.DuplicateName("create_data_array")
        if compression == Compression.Auto:
            compression = self._compr
        da = DataArray._create_new(self, data_arrays, name, array_type,
                                   dtype, shape, compression)
        if data is not None:
            da.write_direct(data)
        return da

    def create_data_frame(self, name="", type_="", col_dict=None,
                          col_names=None, col_dtypes=None, data=None,
                          compression=Compression.No,
                          copy_from=None, keep_copy_id=True):
        """
        Create/copy a new data frame for this block. Either ``col_dict``
        or ``col_name`` and ``col_dtypes`` must be given.
        If both are given, ``col_dict`` will be used.

        :param name: The name of the data frame to create/copy.
        :type name: str
        :param type_: The type of the data frame.
        :type type_: str
        :param col_dict: The dictionary that specifies column
                         names and data type in each column
        :type col_dict: dict or OrderedDict of {str: type}
        :param col_names: The collection of name of all columns in order
        :type col_names: tuples or list or np.array of string
        :param col_dtypes: The collection of data type of all columns in order
        :type col_dtypes: tuples or list or np.array of type
        :param data: Data to write after storage has been created
        :type data: array-like data with compound data type
                    as specified in the columns
        :param compression: En-/disable dataset compression.
        :type compression: :class:`~nixio.Compression`
        :param copy_from: The DataFrame to be copied, None in normal mode
        :type copy_from: DataFrame
        :param keep_copy_id: Specify if the id should be copied in copy mode
        :type keep_copy_id: bool

        :returns: The newly created data frame.
        :rtype: :class:`~nixio.DataFrame`
        """
        if copy_from:
            if not isinstance(copy_from, DataFrame):
                raise TypeError("Object to be copied is not a DataFrame")
            id = self._copy_objects(copy_from, "data_frames",
                                    keep_copy_id, name)
            return self.data_frames[id]

        util.check_entity_name_and_type(name, type_)
        if (isinstance(col_dict, dict)
                and not isinstance(col_dict, OrderedDict)
                and sys.version_info[0] < 3):
            raise TypeError("Python 2 users should use name_list "
                            "or OrderedDict created with LIST and TUPLES "
                            "to create DataFrames as the order "
                            "of the columns cannot be maintained in Py2")

        if data is not None:
            shape = len(data)
        else:
            shape = 0
        data_frames = self._h5group.open_group("data_frames")

        if col_dict is None:
            if col_names is not None:
                if col_dtypes is not None:
                    col_dict = OrderedDict(
                        (str(nam), dt)
                        for nam, dt in zip(col_names, col_dtypes)
                    )
                elif col_dtypes is None and data is not None:
                    col_dtypes = []
                    for x in data[0]:
                        col_dtypes.append(type(x))
                    col_dict = OrderedDict(
                        (str(nam), dt)
                        for nam, dt in zip(col_names, col_dtypes)
                    )
                else:  # col_dtypes is None and data is None
                    raise ValueError(
                           "The data type of each column have to be specified"
                    )
            else:  # if col_names is None
                if data is not None and type(data[0]) == np.void:
                    col_dtype = data[0].dtype
                    for i, dt in enumerate(col_dtype.fields.values()):
                        if dt[0] == np.dtype(str):
                            cn = list(col_dtype.fields.keys())
                            raw_dt = col_dtype.fields.values()
                            raw_dt = list(raw_dt)
                            raw_dt_list = [ele[0] for ele in raw_dt]
                            col_dict = OrderedDict(zip(cn, raw_dt_list))

                else:
                    # data is None or type(data[0]) != np.void
                    # data_type doesnt matter
                    raise ValueError(
                           "No information about column names is provided!"
                    )

        if col_dict is not None:
            for nam, dt in col_dict.items():
                if isclass(dt):
                    if any(issubclass(dt, st) for st in string_types) \
                            or issubclass(dt, np.string_):
                        col_dict[nam] = util.vlen_str_dtype
            dt_arr = list(col_dict.items())
            col_dtype = np.dtype(dt_arr)

        df = DataFrame._create_new(self, data_frames, name,
                                   type_, shape, col_dtype, compression)

        if data is not None:
            if type(data[0]) == np.void:
                data = np.ascontiguousarray(data, dtype=col_dtype)
                df.write_direct(data)
            else:
                data = list(map(tuple, data))
                arr = np.ascontiguousarray(data, dtype=col_dtype)
                df.write_direct(arr)
        return df

    def find_sources(self, filtr=lambda _: True, limit=None):
        """
        Get all sources in this block recursively.

        This method traverses the tree of all sources in the block. The
        traversal is accomplished via breadth first and can be limited in
        depth.  On each node or source a filter is applied. If the filter
        returns true the respective source will be added to the result list.
        By default a filter is used that accepts all sources.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sources.
        :rtype: list of Source
        """
        if limit is None:
            limit = maxint
        return finders._find_sources(self, filtr, limit)

    def pprint(self, indent=2, max_length=120, extra=True, start_depth=0):
        """
        Pretty Printing the Data and MetaData Tree of the whole File

        :param indent: The length of one indentation space
        :type indent: int
        :param max_length: Maximum length of each line of output
        :type max_length: int
        :param extra: True to print extra information of Entities
        :type extra: bool
        :param start_depth: Starting depth of indentation
        :type start_depth: int
        """
        print("{}{}".format(" " * indent*start_depth, self))
        for grp in self.groups:
            self._pp(grp, max_length, indent*(start_depth + 1), False)
            for da in grp.data_arrays:
                self._pp(da, max_length, indent*(start_depth + 2),
                         extra, True)
                for dim in da.dimensions:
                    self._pp(dim, max_length, indent*(start_depth + 3), False)
            for df in grp.data_frames:
                self._pp(df, max_length,
                         indent*(start_depth + 2), extra, True)
            for tag in grp.tags:
                self._pp(tag, max_length,
                         indent*(start_depth + 2), extra, True)
                for fe in tag.features:
                    self._pp(fe, max_length, indent*(start_depth + 3), False)
            for mt in grp.multi_tags:
                self._pp(mt, max_length, indent*(start_depth + 2), extra, True)
                for fe in mt.features:
                    self._pp(fe, max_length, indent*(start_depth + 3), False)
        for da in self.data_arrays:
            self._pp(da, max_length, indent*(start_depth + 1), extra)
            for dim in da.dimensions:
                self._pp(dim, max_length, indent*(start_depth + 2), False)
        for df in self.data_frames:
            self._pp(df, max_length, indent*(start_depth + 1), extra)
        for tag in self.tags:
            self._pp(tag, max_length, indent*(start_depth + 1), extra)
            for fe in tag.features:
                self._pp(fe, max_length, indent*(start_depth + 2), False)
        for mt in self.multi_tags:
            self._pp(mt, max_length, indent*(start_depth + 1), extra)
            for fe in mt.features:
                self._pp(fe, max_length, indent*(start_depth + 2), False)

    @staticmethod
    def _pp(obj, ml, indent, ex, grp=False):
        spaces = " "*(indent)
        if grp:
            prefix = "*"
        else:
            prefix = ""
        if ex:
            stat = ""
            if isinstance(obj, MultiTag):
                stat = "Position Shape:{} Units: {}".format(
                    obj.positions.shape,
                    obj.units)
            elif isinstance(obj, Tag):
                stat = "Position Length:{} Units: {}".format(len(obj.position),
                                                             obj.units)
            elif isinstance(obj, DataFrame):
                stat = "Shape: {} Columns:{}".format(obj.shape,
                                                     obj.column_names)
            elif isinstance(obj, DataArray):
                stat = "Shape: {} Unit:{}".format(obj.shape, obj.unit)
            p = "{}{}{}".format(spaces, prefix, obj)
            n = "{}  {}".format(spaces, stat)
        else:
            p = "{}{}{}".format(spaces, prefix, obj)
        if len(p) > ml - 4:
            split_len = int(ml/2)
            str1 = p[0:split_len]
            str2 = p[-split_len:]
            print("{} ... {}".format(str1, str2))
        else:
            print(p)
        if ex:
            if len(n) > ml - 4:
                split_len = int(ml/2)
                nstr1 = n[0:split_len]
                nstr2 = n[-split_len:]
                print("{} ... {}".format(nstr1, nstr2))
            else:
                print(n)

    def _copy_objects(self, obj, clsname, keep_id=True, name=""):
        src = "{}/{}".format(clsname, obj.name)
        if not name:
            name = str(obj.name)
        ogrp = self._h5group.open_group(clsname, True)
        if name in ogrp:
            raise NameError("Name already exist. Possible solution is to "
                            "provide a new name when copying destination "
                            "is the same as the source parent")
        o = obj._parent._h5group.copy(source=src, dest=self._h5group,
                                      name=name, cls=clsname,
                                      keep_id=keep_id)

        return o.attrs["entity_id"]

    @property
    def sources(self):
        """
        A property containing all sources of a block. Sources can be obtained
        via their index or by their id. Sources can be deleted from the list.
        Adding sources is done using the Blocks create_source method.
        This is a read only attribute.
        """
        if self._sources is None:
            self._sources = SourceContainer("sources", self, Source)
        return self._sources

    @property
    def multi_tags(self):
        """
        A property containing all multi tags of a block. MultiTag entities can
        be obtained via their index or by their id. Tags can be deleted from
        the list. Adding tags is done using the Blocks create_multi_tag method.
        This is a read only attribute.
        """
        if self._multi_tags is None:
            self._multi_tags = Container("multi_tags", self, MultiTag)
        return self._multi_tags

    @property
    def tags(self):
        """
        A property containing all tags of a block. Tag entities can be obtained
        via their index or by their id. Tags can be deleted from the list.
        Adding tags is done using the Blocks create_tag method.
        This is a read only attribute.
        """
        if self._tags is None:
            self._tags = Container("tags", self, Tag)
        return self._tags

    @property
    def data_arrays(self):
        """
        A property containing all data arrays of a block. DataArray entities
        can be obtained via their index or by their id. Data arrays can be
        deleted from the list. Adding a data array is done using the Blocks
        create_data_array method.
        This is a read only attribute.
        """
        if self._data_arrays is None:
            self._data_arrays = Container("data_arrays", self, DataArray)
        return self._data_arrays

    @property
    def data_frames(self):
        if self._data_frames is None:
            self._data_frames = Container("data_frames", self, DataFrame)
        return self._data_frames

    @property
    def groups(self):
        """
        A property containing all groups of a block. Group entities can be
        obtained via their index or by their id. Groups can be deleted from the
        list. Adding a Group is done using the Blocks create_group method.
        This is a read only attribute.
        """
        if self._groups is None:
            self._groups = Container("groups", self, Group)
        return self._groups

    # metadata
    @property
    def metadata(self):
        """
        Associated metadata of the entity. Sections attached to the entity via
        this attribute can provide additional annotations. This is an optional
        read-write property, and can be None if no metadata is available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("{} is not of type Section".format(sect))
        self._h5group.create_link(sect, "metadata")

    @metadata.deleter
    def metadata(self):
        if "metadata" in self._h5group:
            self._h5group.delete("metadata")



# -*- coding: utf-8 -*-
# Copyright © 2019, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)
from .util import units
from .dimension_type import DimensionType


class ValidationError(object):
    """
    Static class for defining validation error messages
    """
    NoName = "no name set"
    NoType = "no type set"
    NoDate = "date is not set"
    NoDataType = "data type is not set"
    NoID = "no ID set"
    DimensionMismatch = ("data dimensionality does not match number of "
                         "defined dimensions")
    InvalidDimensionIndex = ("index for dimension {} is not set to a valid "
                             "value (index > 0)")
    IncorrectDimensionIndex = ("index for dimension {} is set to incorrect "
                               "value {}")
    DimensionTypeMismatch = ("dimension_type attribute for dimension "
                             "{} does not match Dimension object type")
    RangeDimTicksMismatch = ("number of ticks in RangeDimension ({}) "
                             "differs from the number of data entries "
                             "along the corresponding data dimension")
    SetDimLabelsMismatch = ("number of labels in SetDimension ({}) "
                            "differs from the number of data entries "
                            "along the corresponding data dimension")
    NoPosition = "position is not set"
    PositionDimensionMismatch = ("number of entries in position does not "
                                 "match number of dimensions in all "
                                 "referenced DataArrays")
    ExtentDimensionMismatch = ("number of entries in extent does not match "
                               "number of dimensions in all referenced "
                               "DataArrays")
    PositionExtentMismatch = ("number of entries in position and extent "
                              "do not match")
    ReferenceUnitsMismatch = ("some of the referenced DataArrays' dimensions "
                              "don't have units where the Tag has; "
                              "make sure that all references have the same "
                              "number of dimensions as the Tag has units "
                              "and that each dimension has a unit set")
    ReferenceUnitsIncompatible = ("some of the referenced DataArrays' "
                                  "dimensions have units that are not "
                                  "convertible to the units set in the Tag "
                                  "(Note: composite units are not supported)")
    InvalidUnit = ("unit is invalid: not an atomic SI "
                   "(Note: composite units are not supported)")
    NoPositions = "positions are not set"
    PositionsDimensionMismatch = ("number of entries (in 2nd dim) in "
                                  "positions does not match number of "
                                  "dimensions in all referenced DataArrays")
    ExtentsDimensionMismatch = ("number of entries (in 2nd dim) in extents "
                                "does not match number of dimensions in all "
                                "referenced DataArrays")
    PositionsExtentsMismatch = ("number of entries in positions and extents "
                                "do not match")
    NoTicks = "ticks for dimension {} are not set"
    UnsortedTicks = "ticks for dimension {} are not sorted"
    InvalidDimensionUnit = ("unit for dimension {} is set but it is not an "
                            "atomic SI unit "
                            "(Note: composite units are not supported)")
    NoSamplingInterval = ("sampling interval for dimension {} "
                          "is not set")
    InvalidSamplingInterval = ("sampling interval for dimension {} "
                               "is not valid (interval > 0)")
    NoData = "data is not set"
    NoLinkType = "link_type is not set"


class ValidationWarning(object):
    NoVersion = "version is not set"
    NoFormat = "format is not set"
    InvalidUnit = "unit is not SI or composite of SI units"
    NoExpansionOrigin = ("polynomial coefficients for calibration are set, "
                         "but expansion origin is missing")
    NoPolynomialCoefficients = ("expansion origin for calibration is set, "
                                "but polynomial coefficients are missing")
    OffsetNoUnit = ("offset for dimension {} is set, "
                    "but no valid unit is set")
    NoUnit = "unit is not set"


def check_file(nixfile):
    """
    Validate a NIX file and all contained objects and return all errors and
    warnings for each individual object.

    :returns: A nested dictionary of errors and warnings. Each subdictionary
    is indexed by the object with values describing the error or warning.
    :rtype: Dictionary
    """
    results = {"errors": dict(), "warnings": dict()}
    if not nixfile.created_at:
        results["errors"][nixfile] = [ValidationError.NoDate]

    file_warnings = list()
    if not nixfile.version:
        file_warnings.append(ValidationWarning.NoVersion)
    if not nixfile.format:
        file_warnings.append(ValidationWarning.NoFormat)
    if file_warnings:
        results["warnings"][nixfile] = file_warnings

    def update_results(obj, errors, warnings):
        if errors:
            results["errors"][obj] = errors
        if warnings:
            results["warnings"][obj] = warnings

    # Sources
    def traverse_sources(sources):
        # for recursively checking a source tree
        for source in sources:
            src_errors, src_warnings = check_source(source)
            update_results(source, src_errors, src_warnings)
            traverse_sources(source.sources)

    # Sections
    def traverse_sections(sections):
        # for recursively checking a metadata tree
        for section in sections:
            sec_errors, sec_warnings = check_section(section)
            update_results(section, sec_errors, sec_warnings)
            traverse_sections(section.sections)

    # Blocks
    for block in nixfile.blocks:
        blk_errors, blk_warnings = check_block(block)
        update_results(block, blk_errors, blk_warnings)

        # Groups
        for group in block.groups:
            grp_errors, grp_warnings = check_group(group)
            update_results(group, grp_errors, grp_warnings)

        # DataArrays
        for da in block.data_arrays:
            da_errors, da_warnings = check_data_array(da)
            update_results(da, da_errors, da_warnings)

        # Tags
        for tag in block.tags:
            tag_errors, tag_warnings = check_tag(tag)
            update_results(tag, tag_errors, tag_warnings)

        # MultiTags
        for mtag in block.multi_tags:
            mtag_errors, mtag_warnings = check_multi_tag(mtag)
            update_results(mtag, mtag_errors, mtag_warnings)

        traverse_sources(block.sources)

    traverse_sections(nixfile.sections)

    return results


def check_block(block):
    """
    Validate a Block and return all errors and warnings for the block and each
    individual contained object.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(block)
    return errors, list()


def check_group(group):
    """
    Validate a Group and return all errors and warnings.
    Does not check contained objects.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(group)
    return errors, list()


def check_data_array(da):
    """
    Validate a DataArray and its Dimensions and return all errors and warnings.
    Errors and warnings about Dimension objects are included in the DataArray
    errors and warnings lists.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(da)
    warnings = list()

    if not da.data_type:
        errors.append(ValidationError.NoDataType)

    if len(da.dimensions) != len(da.shape):
        errors.append(ValidationError.DimensionMismatch)

    if da.unit and not units.is_si(da.unit):
        warnings.append(ValidationWarning.InvalidUnit)

    if da.polynom_coefficients and not da.expansion_origin:
        warnings.append(ValidationWarning.NoExpansionOrigin)
    elif da.expansion_origin and not da.polynom_coefficients:
        warnings.append(ValidationWarning.NoPolynomialCoefficients)

    for idx, (dim, datalen) in enumerate(zip(da.dimensions, da.shape), 1):
        if not dim.index or dim.index <= 0:
            errors.append(ValidationError.InvalidDimensionIndex.format(idx))
        elif dim.index != idx:
            errors.append(
                ValidationError.IncorrectDimensionIndex.format(idx, dim.index)
            )
        if dim.dimension_type == DimensionType.Range:
            if dim.ticks is not None and len(dim.ticks) != datalen:
                # if ticks is None or empty, it will be reported by the
                # dimension check function
                errors.append(
                    ValidationError.RangeDimTicksMismatch.format(idx)
                )
            dim_errors, dim_warnings = check_range_dimension(dim, idx)
        elif dim.dimension_type == DimensionType.Sample:
            dim_errors, dim_warnings = check_sampled_dimension(dim, idx)
        elif dim.dimension_type == DimensionType.Set:
            if dim.labels and len(dim.labels) != datalen:
                # empty labels is allowed
                errors.append(ValidationError.SetDimLabelsMismatch.format(idx))
            dim_errors, dim_warnings = check_set_dimension(dim, idx)
        errors.extend(dim_errors)
        warnings.extend(dim_warnings)
    return errors, warnings


def check_tag(tag):
    """
    Validate a Tag and its Features and return all errors and warnings.
    Errors and warnings about Features are included in the Tag errors and
    warnings lists.

    For Features, only the basic Entity requirements, LinkType, and the
    existence of a referenced DataArray are checked.  The linked DataArray is
    not checked for validity (see check_data_array() for those validation
    checks).

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(tag)
    warnings = list()

    if not tag.position:
        errors.append(ValidationError.NoPosition)
    if tag.references:
        posdim = len(tag.position)
        if any(posdim != len(da.shape) for da in tag.references):
            errors.append(ValidationError.PositionDimensionMismatch)
        if tag.extent:
            extlen = len(tag.extent)
            if extlen != posdim:
                errors.append(ValidationError.PositionExtentMismatch)
            if any(extlen != len(da.shape) for da in tag.references):
                errors.append(ValidationError.ExtentDimensionMismatch)

        refs_units = [get_dim_units(da) for da in tag.references]
        if any(len(ru) != len(tag.units) for ru in refs_units):
            errors.append(ValidationError.ReferenceUnitsMismatch)

        if not tag_units_match_refs_units(tag.units, refs_units):
            errors.append(ValidationError.ReferenceUnitsIncompatible)

    if any(not units.is_si(u) for u in tag.units if u):
        errors.append(ValidationError.InvalidUnit)

    for idx, feat in enumerate(tag.features):
        feat_errors = check_feature(feat, idx)
        errors.extend(feat_errors)

    return errors, warnings


def check_multi_tag(mtag):
    """
    Validate a MultiTag and its Features and return all errors and warnings.
    Errors and warnings about Features are included in the MultiTag errors and
    warnings lists.

    For Features, only the basic Entity requirements, LinkType, and the
    existence of a referenced DataArray are checked.  The linked DataArray is
    not checked for validity (see check_data_array() for those validation
    checks).

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(mtag)
    warnings = list()

    if not mtag.positions:
        errors.append(ValidationError.NoPositions)
    if mtag.references:
        if len(mtag.positions.shape) == 1:
            posdim = 1
        else:
            posdim = mtag.positions.shape[1]
        # New error for len(mtag.positions.shape) > 2
        if any(posdim != len(da.shape) for da in mtag.references):
            errors.append(ValidationError.PositionsDimensionMismatch)
        if mtag.extents:
            if mtag.positions.shape != mtag.extents.shape:
                errors.append(ValidationError.PositionsExtentsMismatch)
            if len(mtag.extents.shape) == 1:
                extdim = 1
            else:
                extdim = mtag.extents.shape[1]
            if any(extdim != len(da.shape) for da in mtag.references):
                errors.append(ValidationError.ExtentsDimensionMismatch)

        refs_units = [get_dim_units(da) for da in mtag.references]
        if any(len(ru) != len(mtag.units) for ru in refs_units):
            errors.append(ValidationError.ReferenceUnitsMismatch)

        if not tag_units_match_refs_units(mtag.units, refs_units):
            errors.append(ValidationError.ReferenceUnitsIncompatible)

    if any(not units.is_si(u) for u in mtag.units if u):
        errors.append(ValidationError.InvalidUnit)

    for idx, feat in enumerate(mtag.features):
        feat_errors = check_feature(feat, idx)
        errors.extend(feat_errors)

    return errors, warnings


def check_feature(feat, idx):
    """
    Validate a Feature and return all errors.

    :returns: A list of 'errors'
    """
    errors = list()
    if not feat.id:
        errors.append("feature {}: {}".format(idx, ValidationError.NoID))
    if feat.created_at is None:
        errors.append("feature {}: {}".format(idx, ValidationError.NoDate))
    if not feat.data:
        errors.append("feature {}: {}".format(idx, ValidationError.NoData))
    if not feat.link_type:
        errors.append("feature {}: {}".format(idx, ValidationError.NoLinkType))
    return errors


def check_section(section):
    """
    Validate a Section and its Properties return all errors and warnings.

    Errors and warnings about Properties are included in the Section errors and
    warnings lists.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(section)
    warnings = list()
    for idx, prop in enumerate(section.props):
        prop_errors, prop_warnings = check_property(prop, idx)
        errors.extend(prop_errors)
        warnings.extend(prop_warnings)

    return errors, warnings


def check_property(prop, idx):
    """
    Validate a Property and return all errors.

    :returns: A list of 'errors' and 'warnings'
    """
    errors = list()
    warnings = list()
    if not prop.id:
        errors.append("property {}: {}".format(idx, ValidationError.NoID))
    if not prop.name:
        errors.append("property {}: {}".format(idx, ValidationError.NoName))
    if not prop.unit:
        warnings.append("property {}: {}".format(idx,
                                                 ValidationWarning.NoUnit))
    return errors, warnings


def check_source(source):
    """
    Validate a Source and return all errors and warnings.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = check_entity(source)
    return errors, list()


def check_range_dimension(dim, idx):
    """
    Validate a RangeDimension and return all errors and warnings.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = list()
    warnings = list()

    if not dim.ticks:
        errors.append(ValidationError.NoTicks.format(idx))
    elif not all(ti < tj for ti, tj in zip(dim.ticks[:-1], dim.ticks[1:])):
        errors.append(ValidationError.UnsortedTicks.format(idx))

    if dim.unit and not units.is_atomic(dim.unit):
        errors.append(ValidationError.InvalidDimensionUnit.format(idx))
    return errors, warnings


def check_set_dimension(dim, idx):
    """
    Validate a SetDimension and return all errors and warnings.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    return list(), list()


def check_sampled_dimension(dim, idx):
    """
    Validate a SetDimension and return all errors and warnings.

    :returns: A list of 'errors' and a list of 'warnings'
    """
    errors = list()
    warnings = list()

    if not dim.sampling_interval:
        errors.append(ValidationError.NoSamplingInterval.format(idx))
    elif dim.sampling_interval < 0:
        errors.append(ValidationError.InvalidSamplingInterval.format(idx))

    if dim.unit:
        if not units.is_atomic(dim.unit):
            errors.append(ValidationError.InvalidDimensionUnit.format(idx))
    else:
        if dim.offset:
            warnings.append(ValidationWarning.OffsetNoUnit.format(idx))
    return errors, warnings


def check_entity(entity):
    """
    General NIX entity validator
    """
    errors = []
    if not entity.type:
        errors.append(ValidationError.NoType)
    if not entity.id:
        errors.append(ValidationError.NoID)
    if not entity.name:
        errors.append(ValidationError.NoName)
    if entity.created_at is None:
        errors.append(ValidationError.NoDate)
    return errors


def get_dim_units(data_array):
    """
    Helper function to collect the units of the dimensions of a data array
    """
    unit_list = []
    for dim in data_array.dimensions:
        if (dim.dimension_type == DimensionType.Range or
                dim.dimension_type == DimensionType.Sample):
            unit_list.append(dim.unit if dim.unit else "")
        elif dim.dimension_type == DimensionType.Set:
            unit_list.append("")
    return unit_list


def tag_units_match_refs_units(tag_units, refs_units):
    for ref in refs_units:
        for ru, tu in zip(tag_units, ref):
            if ru == "" and tu == "":
                continue
            if not units.scalable(ru, tu):
                return False
    return True



import sys
import os
import nixio as nix


def usage():
    print("Usage")
    print("  nix-validate <nixfile>...")
    print()
    print("Args")
    print("  <nixfile>...    One or more NIX files")


def format_obj(obj):
    return "{} '{}' (ID: {})".format(type(obj).__name__, obj.name, obj.id)


def print_header(text, h=1):
    print("  {}".format(text))
    print("  " + "-"*len(text))


def validate(filename):
    nf = nix.File(filename, mode=nix.FileMode.ReadOnly)
    results = nf.validate()
    print("Results for '{}'".format(filename))
    errors = results["errors"]
    if errors:
        plural = "s" if len(errors) > 1 else ""
        print_header("{} object{} with errors".format(len(errors), plural))
        for idx, (obj, messages) in enumerate(errors.items()):
            print(" [{}] {}".format(idx+1, format_obj(obj)))
            for msg in messages:
                print("    {}".format(msg))

    warnings = results["warnings"]
    if warnings:
        plural = "s" if len(warnings) > 1 else ""
        print_header("{} object{} with warnings".format(len(warnings), plural))
        for idx, (obj, messages) in enumerate(warnings.items()):
            print(" [{}] {}".format(idx+1, format_obj(obj)))
            for msg in messages:
                print("    {}".format(msg))

    print()


def main():
    args = sys.argv
    if len(args) < 2:
        usage()
        sys.exit(1)

    nixfnames = args[1:]
    for nixfn in nixfnames:
        if os.path.exists(nixfn):
            validate(nixfn)
        else:
            print("error: No such file '{}'".format(nixfn))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
import warnings
from .tag import BaseTag, FeatureContainer
from .container import LinkContainer
from .feature import Feature
from .source_link_container import SourceLinkContainer
from .data_array import DataArray
from .data_view import DataView
from .link_type import LinkType
from .exceptions import (OutOfBounds, IncompatibleDimensions,
                         UninitializedEntity)
from .section import Section


class MultiTag(BaseTag):

    def __init__(self, nixparent, h5group):
        super(MultiTag, self).__init__(nixparent, h5group)
        self._sources = None
        self._references = None
        self._features = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, positions):
        newentity = super(MultiTag, cls)._create_new(nixparent, h5parent,
                                                     name, type_)
        newentity.positions = positions
        return newentity

    @property
    def positions(self):
        """
        The positions defined by the tag. This is a read-write property.

        :type: DataArray
        """
        if "positions" not in self._h5group:
            raise RuntimeError("MultiTag.positions: DataArray not found!")
        return DataArray(self._parent, self._h5group.open_group("positions"))

    @positions.setter
    def positions(self, da):
        if da is None:
            raise TypeError("MultiTag.positions cannot be None.")
        if "positions" in self._h5group:
            del self._h5group["positions"]
        self._h5group.create_link(da, "positions")
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    @property
    def extents(self):
        """
        The extents defined by the tag. This is an optional read-write
        property and may be set to None.

        :type: DataArray or None
        """
        if "extents" in self._h5group:
            return DataArray(self._parent, self._h5group.open_group("extents"))
        else:
            return None

    @extents.setter
    def extents(self, da):
        if da is None:
            del self._h5group["extents"]
        else:
            self._h5group.create_link(da, "extents")
        if self._parent._parent.time_auto_update:
            self.force_updated_at()

    @property
    def references(self):
        """
        A property containing all data arrays referenced by the tag. Referenced
        data arrays can be obtained by index or their id. References can be
        removed from the list, removing a referenced DataArray will not remove
        it from the file. New references can be added using the append method
        of the list.
        This is a read only attribute.

        :type: LinkContainer of DataArray
        """
        if self._references is None:
            self._references = LinkContainer("references", self, DataArray,
                                             self._parent.data_arrays)
        return self._references

    @property
    def features(self):
        """
        A property containing all features of the tag. Features can be obtained
        via their index or their id. Features can be deleted from the list.
        Adding new features to the multitag is done using the create_feature
        method. This is a read only attribute.

        :type: Container of Feature.
        """
        if self._features is None:
            self._features = FeatureContainer("features", self, Feature)
        return self._features

    def _get_slice(self, data, index):
        offset, count = self._get_offset_and_count(data, index)
        sl = tuple(slice(o, o+c) for o, c in zip(offset, count))
        return sl

    def _calc_data_slices(self, data, index):
        positions = self.positions
        extents = self.extents

        pos_size = positions.data_extent if positions else tuple()
        ext_size = extents.data_extent if extents else tuple()

        if not positions or index >= pos_size[0]:
            raise OutOfBounds("Index out of bounds of positions!")

        if extents and index >= ext_size[0]:
            raise OutOfBounds("Index out of bounds of extents!")

        incdim_exception = IncompatibleDimensions(
                "Number of dimensions in positions does not match "
                "dimensionality of data",
                "MultiTag._calc_data_slices"
            )

        if len(pos_size) == 1 and len(data.dimensions) != 1:
            raise incdim_exception

        if len(pos_size) > 1 and pos_size[1] > len(data.dimensions):
            raise incdim_exception

        if (extents and len(ext_size) > 1 and
                ext_size[1] > len(data.dimensions)):
            raise incdim_exception

        if len(pos_size) == 1:
            dimpos = positions[0:len(data.dimensions)]
        else:
            dimpos = positions[index, 0:len(data.dimensions)]
        units = self.units
        starts, stops = list(), list()
        for idx in range(dimpos.size):
            dim = data.dimensions[idx]
            unit = None
            if idx <= len(units) and len(units):
                unit = units[idx]
            starts.append(self._pos_to_idx(dimpos.item(idx), unit, dim))

        if extents:
            extent = extents[index, 0:len(data.dimensions)]
            for idx in range(extent.size):
                dim = data.dimensions[idx]
                unit = None
                if idx <= len(units) and len(units):
                    unit = units[idx]
                stop = self._pos_to_idx(dimpos.item(idx) + extent[idx],
                                        unit, dim)
                stop += 1
                minstop = starts[idx] + 1
                stops.append(max(stop, minstop))
        else:
            stops = [start+1 for start in starts]

        return tuple(slice(start, stop) for start, stop in zip(starts, stops))

    def retrieve_data(self, posidx, refidx):
        msg = ("Call to deprecated method MultiTag.retrieve_data. "
               "Use MultiTag.tagged_data instead.")
        warnings.warn(msg, category=DeprecationWarning)
        return self.tagged_data(posidx, refidx)

    def tagged_data(self, posidx, refidx):
        references = self.references
        positions = self.positions
        extents = self.extents
        if len(references) == 0:
            raise OutOfBounds("There are no references in this multitag!")

        if (posidx >= positions.data_extent[0] or
                extents and posidx >= extents.data_extent[0]):
            raise OutOfBounds("Index out of bounds of positions or extents!")

        ref = references[refidx]
        dimcount = len(ref.dimensions)
        if len(positions.data_extent) == 1 and dimcount != 1:
            raise IncompatibleDimensions(
                "Number of dimensions in position or extent do not match "
                "dimensionality of data",
                "MultiTag.tagged_data")
        if len(positions.data_extent) > 1:
            if (positions.data_extent[1] > dimcount or
                    extents and extents.data_extent[1] > dimcount):
                raise IncompatibleDimensions(
                    "Number of dimensions in position or extent do not match "
                    "dimensionality of data",
                    "MultiTag.tagged_data")
        slices = self._calc_data_slices(ref, posidx)

        if not self._slices_in_data(ref, slices):
            raise OutOfBounds("References data slice out of the extent of the "
                              "DataArray!")
        return DataView(ref, slices)

    def retrieve_feature_data(self, posidx, featidx):
        msg = ("Call to deprecated method MultiTag.retrieve_feature_data. "
               "Use MultiTag.feature_data instead.")
        warnings.warn(msg, category=DeprecationWarning)
        return self.feature_data(posidx, featidx)

    def feature_data(self, posidx, featidx):
        if len(self.features) == 0:
            raise OutOfBounds(
                "There are no features associated with this tag!"
            )

        try:
            feat = self.features[featidx]
        except KeyError:
            feat = None
            for f in self.features:
                if f.data.name == featidx or f.data.id == featidx:
                    feat = f
                    break
            if feat is None:
                raise
        da = feat.data
        if da is None:
            raise UninitializedEntity()
        if feat.link_type == LinkType.Tagged:
            slices = self._calc_data_slices(da, posidx)
            if not self._slices_in_data(da, slices):
                raise OutOfBounds("Requested data slice out of the extent "
                                  "of the Feature!")
            return DataView(da, slices)
        elif feat.link_type == LinkType.Indexed:
            if posidx > da.data_extent[0]:
                raise OutOfBounds("Position is larger than the data stored "
                                  "in the Feature!")
            slices = [slice(posidx, posidx+1)]
            slices.extend(slice(0, stop) for stop in da.data_extent[1:])

            if not self._slices_in_data(da, slices):
                OutOfBounds("Requested data slice out of the extent of the "
                            "Feature!")
            return DataView(da, slices)
        # For untagged return the full data
        slices = tuple(slice(0, stop) for stop in da.data_extent)
        return DataView(da, slices)

    @property
    def sources(self):
        """
        A property containing all Sources referenced by the MultiTag. Sources
        can be obtained by index or their id. Sources can be removed from the
        list, but removing a referenced Source will not remove it from the
        file. New Sources can be added using the append method of the list.
        This is a read only attribute.
        """
        if self._sources is None:
            self._sources = SourceLinkContainer(self)
        return self._sources

    # metadata
    @property
    def metadata(self):
        """

        Associated metadata of the entity. Sections attached to the entity via
        this attribute can provide additional annotations. This is an optional
        read-write property, and can be None if no metadata is available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("{} is not of type Section".format(sect))
        self._h5group.create_link(sect, "metadata")

    @metadata.deleter
    def metadata(self):
        if "metadata" in self._h5group:
            self._h5group.delete("metadata")



# -*- coding: utf-8 -*-
# Copyright © 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

import h5py
import numpy as np

from .h5dataset import H5DataSet
from ..datatype import DataType
from ..block import Block
from ..section import Section

from .. import util
from ..exceptions import InvalidEntity


class H5Group(object):

    def __init__(self, parent, name, create=False):
        self._parent = parent
        self.name = name
        self.group = None
        if create or name in self._parent:
            self._create_h5obj()
        self.h5obj = self.group

    def _create_h5obj(self):
        if self.name in self._parent:
            self.group = self._parent[self.name]
        else:
            gcpl = h5py.h5p.create(h5py.h5p.GROUP_CREATE)
            flags = h5py.h5p.CRT_ORDER_TRACKED | h5py.h5p.CRT_ORDER_INDEXED
            gcpl.set_link_creation_order(flags)
            name = self.name.encode("utf-8")
            gid = h5py.h5g.create(self._parent.id, name, gcpl=gcpl)
            self.group = h5py.Group(gid)

    @property
    def group(self):
        if self._group is None:
            if self.name in self._parent:
                self._group = self._parent[self.name]
            else:
                return None
        return self._group

    @group.setter
    def group(self, grp):
        self._group = grp

    def create_link(self, target, name):
        self._create_h5obj()
        if name in self.group:
            del self.group[name]
        self.group[name] = target._h5group.group

    @classmethod
    def create_from_h5obj(cls, h5obj):
        parent = h5obj.parent
        name = h5obj.name.split("/")[-1]
        if isinstance(h5obj, h5py.Group):
            return cls(parent, name)
        elif isinstance(h5obj, h5py.Dataset):
            return H5DataSet(parent, name)
        else:
            raise ValueError("Invalid object: "
                             "{} must be either h5py.Group of h5py.Dataset.")

    def open_group(self, name, create=False):
        """
        Returns a new H5Group with the given name contained in the current
        group.  If the current group does not exist in the file,
        it is automatically created.

        :param name: the name of the group
        :param create: creates the child group in the file if it does not exist
        :return: a new H5Group object
        """
        self._create_h5obj()
        return H5Group(self.group, name, create)

    def create_dataset(self, name, shape, dtype, compression=False):
        """
        Creates a dataset object under the current group with a given name,
        shape, and type.

        :param name: the name of the dataset
        :param shape: tuple representing the shape of the dataset
        :param dtype: the type of the data for this dataset (DataType)
        :param compression: whether to compress the data (default: False)
        :return: a new H5DataSet object
        """
        self._create_h5obj()
        return H5DataSet(self.group, name, dtype, shape, compression)

    def get_dataset(self, name):
        """
        Returns a contained H5DataSet object.

        :param name: name of the dataset
        :return: H5DataSet object
        """
        notfound = KeyError("No DataSet named {} found.")
        if self.group is None:
            raise notfound
        if name in self.group:
            dset = self.group[name]
            return H5DataSet.create_from_h5obj(dset)
        else:
            raise notfound

    def write_data(self, name, data, dtype=None, compression=False):
        """
        Writes the data to a Dataset contained in the group with the
        given name. Creates the Dataset if necessary.

        :param name: name of the Dataset object
        :param data: the data to write
        :param dtype: optionally specify the data type, otherwise it will be
        automatically determined by the data
        :param compression: whether to compress the data (default: False)
        """
        shape = np.shape(data)
        if self.has_data(name):
            dset = self.get_dataset(name)
            dset.shape = shape
        else:
            if dtype is None:
                dtype = DataType.get_dtype(data[0])
            dset = self.create_dataset(name, shape, dtype, compression)

        dset.write_data(data)

    def get_data(self, name):
        """
        Returns the data contained in the dataset identified by 'name', or an
        empty list if a dataset of that name does not exist in the Group.

        :param name: The name of the dataset
        :return: The data contained in the dataset as a numpy array or None
        """
        if name not in self.group:
            return []

        dset = self.group[name]
        # TODO: Error if dset is Group?
        return dset[:]

    def has_data(self, name):
        """
        Return True if the Group contains a Dataset object with the given name.

        :param name: name of Dataset
        :return: True if Dataset exists in Group, False if it does not exist,
        or exists and is not a Dataset
        """
        if self.group.get(name, getclass=True) == h5py.Dataset:
            return True
        else:
            return False

    def has_by_id(self, id_or_name):
        if not self.group:
            return False
        if util.is_uuid(id_or_name):
            for item in self:
                if item.get_attr("entity_id") == id_or_name:
                    return True
            return False
        else:
            return id_or_name in self.group

    def get_by_id_or_name(self, id_or_name):
        if util.is_uuid(id_or_name):
            return self.get_by_id(id_or_name)
        else:
            return self.get_by_name(id_or_name)

    def get_by_name(self, name):
        if self.group and name in self.group:
            return self.create_from_h5obj(self.group[name])
        else:
            raise KeyError("Item not found '{}'".format(name))

    def get_by_id(self, id_):
        if self.group:
            for item in self:
                if item.get_attr("entity_id") == id_:
                    return item
        raise KeyError("Item not found '{}'".format(id_))

    def get_by_pos(self, pos):
        if not self.group:
            raise IndexError

        # Using low level interface to specify iteration order
        name, _ = self.group.id.links.iterate(lambda n: n,
                                              idx_type=h5py.h5.INDEX_CRT_ORDER,
                                              order=h5py.h5.ITER_INC,
                                              idx=pos)
        return self.get_by_name(name)

    def delete(self, id_or_name):
        """
        Deletes the child HDF5 group that matches the given name or id.
        """
        if util.is_uuid(id_or_name):
            name = self.get_by_id_or_name(id_or_name).name
        else:
            name = id_or_name
        try:
            del self.group[name]
        except Exception:
            raise ValueError("Error deleting {} ".format(name))
        # Delete if empty and non-root container
        groupdepth = len(self.group.name.split("/")) - 1
        if not len(self.group) and groupdepth > 1:
            del self.parent.group[self.name]
            # del self.group
            self.group = None

    def delete_all(self, eid):
        """
        Deletes all references to a given list of objects, identified by their
        entity_id, below the current object.
        """
        # Use visit_items to traverse groups and check their children.
        # visit_items visits each item only once, so instead of checking
        # whether each item is the one we're searching for, we check whether
        # it *contains* the one we're searching for
        # We delete the child as soon as we find it; this doesn't cause
        # iteration issues since it's deleted before descending into the
        # children of the current group

        def delete_by_id(name, obj):
            if not isinstance(obj, h5py.Group):
                return
            grp = self.create_from_h5obj(obj)
            for ch in grp:
                if ch.get_attr("entity_id") in eid:
                    del grp[ch.name]

        self._group.visititems(delete_by_id)

    def set_attr(self, name, value):
        self._create_h5obj()
        if value is None:
            if name in self.group.attrs:
                del self.group.attrs[name]
        else:
            self.group.attrs[name] = value

    def get_attr(self, name):
        if self.group is None:
            return None
        attr = self.group.attrs.get(name)
        if isinstance(attr, bytes):
            attr = attr.decode()
        return attr

    def find_children(self, filtr=None, limit=None):
        result = []

        def match(name, obj):
            curdepth = len(name.split("/"))
            if limit is not None and curdepth > limit:
                return None

            h5grp = H5Group.create_from_h5obj(obj)
            if filtr is None or filtr(h5grp):
                result.append(h5grp)

        self.group.visititems(match)
        return result

    def copy(self, source, dest, name=None, cls=None, shallow=False,
             keep_id=True):
        grp = self.group
        dest.open_group(cls, create=True)
        dest_grp = dest.group[cls]
        grp.copy(source=source, dest=dest_grp, name=name, shallow=shallow)

        g = dest_grp[name]
        g.attrs["name"] = name
        if not keep_id:
            def change_id(_, igrp):
                if "entity_id" in igrp.attrs:
                    id_ = util.create_id()
                    igrp.attrs.modify("entity_id", np.string_(id_))
            id_ = util.create_id()
            g.attrs.modify("entity_id", np.string_(id_))
            g.visititems(change_id)
        return g

    @property
    def file(self):
        """
        An H5Group object which represents the file root.

        :return: H5Group at '/'
        """
        return H5Group(self.group.file, "/", create=False)

    @property
    def h5root(self):
        """
        Returns the H5Group of the Block or top-level Section which contains
        this object. Returns None if requested on the file root '/' or the
        /data or /metadata groups.

        :return: Top level object containing this group (H5Group)
        """
        pathparts = self.group.name.split("/")
        if len(pathparts) == 3:
            return self
        if self.group.name == "/":
            return None
        if len(pathparts) == 2:
            return None

        return self.parent.h5root

    @property
    def root(self):
        """
        Returns the Block or top-level Section which contains this object.
        Returns None if requested on the file root '/' or the /data or
        /metadata groups.

        :return: Top level object containing this group (Block or Section)
        """
        h5root = self.h5root
        if h5root is None:
            return None
        topgroup = self.group.name.split("/")[1]
        if topgroup == "data":
            cls = Block
        elif topgroup == "metadata":
            cls = Section
        else:
            raise InvalidEntity
        # TODO: Fix this
        return cls(h5root)

    @property
    def parent(self):
        return self.create_from_h5obj(self._parent)

    def __iter__(self):
        if not len(self):
            return
        for grp in self.group.values():
            yield self.create_from_h5obj(grp)

    def __contains__(self, item):
        if self.group is None:
            return False
        return item in self.group

    def __len__(self):
        if self.group is None:
            return 0
        else:
            return len(self.group)

    def __delitem__(self, key):
        del self.group[key]

    def __str__(self):
        return "<H5Group object: {}>".format(self.group.name)



# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
try:
    from sys import maxint
except ImportError:
    from sys import maxsize as maxint
try:
    from collections.abc import Sequence, Iterable
except ImportError:
    from collections import Sequence, Iterable
from six import string_types
import numpy as np
from .container import Container, SectionContainer
from .datatype import DataType
from .entity import Entity
from .property import Property
from .util import find as finders
from . import util
from . import exceptions


class S(object):
    def __init__(self, section_type, section=None):
        self.section_type = section_type
        self.section = section

    def __setitem__(self, key, value):
        self.section[key] = value

    def __setattr__(self, key, value):
        if key in ['section_type', 'section']:
            object.__setattr__(self, key, value)
        else:
            setattr(self.section, key, value)

    def __getattribute__(self, item):
        if item in ['section_type', 'section']:
            return object.__getattribute__(self, item)

        return getattr(self.section, item)


class Section(Entity):

    def __init__(self, nixparent, h5group):
        super(Section, self).__init__(nixparent, h5group)
        self._sec_parent = None
        self._sections = None
        self._properties = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, oid=None):
        newentity = super(Section, cls)._create_new(nixparent, h5parent,
                                                    name, type_)
        if util.is_uuid(oid):
            newentity._h5group.set_attr("entity_id", oid)

        return newentity

    # Section
    def create_section(self, name, type_="undefined", oid=None):
        """
        Creates a new subsection that is a child of this section entity.

        :param name: The name of the section to create.
        :type name: str
        :param type_: The type of the section.
        :type type_: str
        :param oid: object id, UUID string as specified in RFC 4122. If no id
                    is provided, an id will be generated and assigned.
        :type oid: str

        :returns: The newly created section.
        :rtype: Section
        """
        util.check_entity_name_and_type(name, type_)
        sections = self._h5group.open_group("sections", True)
        if name in sections:
            raise exceptions.DuplicateName("create_section")
        sec = Section._create_new(self, sections, name, type_, oid)
        sec._sec_parent = self
        return sec

    # Property
    def create_property(self, name="", values_or_dtype=0, oid=None,
                        copy_from=None, keep_copy_id=True):
        """
        Add a new property to the section.

        :param name: The name of the property to create/copy.
        :type name: str
        :param values_or_dtype: The values of the property or a valid DataType.
        :type values_or_dtype: list of values or a DataType
        :param oid: object id, UUID string as specified in RFC 4122. If no id
                    is provided, an id will be generated and assigned.
        :type oid: str
        :param copy_from: The Property to be copied, None in normal mode
        :type copy_from: Property
        :param keep_copy_id: Specify if the id should be copied in copy mode
        :type keep_copy_id: bool

        :returns: The newly created property.
        :rtype: Property
        """
        if copy_from:
            if not isinstance(copy_from, Property):
                raise TypeError("Object to be copied is not a Property")
            clsname = "properties"
            src = "{}/{}".format(clsname, copy_from.name)
            if not name:
                name = str(copy_from.name)
            properties = self._h5group.open_group("properties", True)
            if name in properties:
                raise NameError("Name already exist. Possible solution is to "
                                "provide a new name when copying destination "
                                "is the same as the source parent")
            p = copy_from._parent._h5group.copy(source=src, dest=self._h5group,
                                                name=name,
                                                cls=clsname,
                                                keep_id=keep_copy_id)

            id_ = p.attrs["entity_id"]
            return self.props[id_]

        vals = values_or_dtype

        properties = self._h5group.open_group("properties", True)
        if name in properties:
            raise exceptions.DuplicateName("create_property")

        # Handle handed in DataType
        if isinstance(vals, type):
            dtype = vals
            vals = []

        # In case of values, make sure boolean value 'False' gets through as
        # well, but ensure that empty values are not allowed, we need a
        # DataType.
        elif vals is None or (isinstance(vals, (Sequence, Iterable)) and
                              not len(vals)):
            raise TypeError(
                "Please provide either a non empty value or a DataType."
            )

        else:
            # Make sure all values are of the same data type
            single_val = vals
            if (isinstance(vals, (Sequence, Iterable)) and
                    not isinstance(vals, string_types)):
                single_val = vals[0]
            else:
                # Make sure the data will always be created with an array.
                vals = [vals]

            # Will raise an error, if the datatype of the first value is not
            # valid.
            dtype = DataType.get_dtype(single_val)

            # Check all values for data type consistency to ensure clean value
            # add. Will raise an exception otherwise.
            for v in vals:
                if DataType.get_dtype(v) != dtype:
                    raise TypeError("Array contains inconsistent values.")

        prop = Property._create_new(self, properties, name, dtype, oid)
        prop.values = vals

        return prop

    def copy_section(self, obj, children=True, keep_id=True, name=""):
        """
        Copy a section to the section.

        :param obj: The Section to be copied
        :type obj: Section
        :param children: Specify if the copy should be recursive
        :type children: bool
        :param keep_id: Specify if the id should be kept
        :type keep_id: bool
        :param name: Name of copied section, Default is name of source section
        :type name: str

        :returns: The copied section
        :rtype: Section
        """
        if not isinstance(obj, Section):
            raise TypeError("Object to be copied is not a Section")

        if obj._sec_parent:
            src = "{}/{}".format("sections", obj.name)
        else:
            src = "{}/{}".format("metadata", obj.name)

        clsname = "sections"
        if not name:
            name = str(obj.name)
        sec = self._h5group.open_group("sections", True)
        if name in sec:
            raise NameError("Name already exist. Possible solution is to "
                            "provide a new name when copying destination "
                            "is the same as the source parent")
        sec = obj._parent._h5group.copy(source=src, dest=self._h5group,
                                        name=name, cls=clsname,
                                        keep_id=keep_id)

        if not children:
            for p in obj.props:
                self.sections[obj.name].create_property(copy_from=p,
                                                        keep_copy_id=keep_id)

        return self.sections[sec.attrs["entity_id"]]

    @property
    def reference(self):
        return self._h5group.get_attr("reference")

    @reference.setter
    def reference(self, ref):
        util.check_attr_type(ref, str)
        self._h5group.set_attr("reference", ref)
        if self.file.time_auto_update:
            self.force_updated_at()

    @property
    def link(self):
        """
        Link to another section. If a section is linked to another section,
        the linking section inherits all properties from the target section.
        This is an optional read-write property and may be set to None.

        :type: Section
        """
        if "link" not in self._h5group:
            return None

        return Section(self, self._h5group.open_group("link"))

    @link.setter
    def link(self, id_or_sec):
        if id_or_sec is None:
            self._h5group.delete("link")
        if isinstance(id_or_sec, Section):
            sec = id_or_sec
        else:
            rootsec = Section(self, self._h5group.h5root)
            sec = rootsec.find_sections(filtr=lambda x: x.id == id_or_sec)

        self._h5group.create_link(sec, "link")
        if self.file.time_auto_update:
            self.force_updated_at()

    def inherited_properties(self):
        properties = self._h5group.open_group("properties")
        inhprops = [Property(self, h5prop) for h5prop in properties]
        if self.link:
            inhprops.extend(self.link.inherited_properties())
        return inhprops

    @property
    def repository(self):
        """
        URL to the terminology repository the section is associated with.
        This is an optional read-write property and may be set to None.

        :type: str
        """
        return self._h5group.get_attr("repository")

    @repository.setter
    def repository(self, r):
        util.check_attr_type(r, str)
        self._h5group.set_attr("repository", r)
        if self.file.time_auto_update:
            self.force_updated_at()

    @property
    def parent(self):
        """
        The parent section. This is a read-only property. For root sections
        this property is always None.

        Accessing this property can be slow when the metadata tree is large.

        :type: Section
        """
        if self._sec_parent is not None:
            return self._sec_parent
        rootmd = self._h5group.file.open_group("metadata")
        # Assuming most metadata trees are shallow---doing BFS
        sections = [Section(None, sg) for sg in rootmd]
        if self in sections:
            # Top-level section
            return None

        while sections:
            sect = sections.pop(0)
            if self in sect.sections:
                self._sec_parent = sect
                return sect
            sections.extend(sect.sections)

        return None

    @property
    def file(self):
        """
        Root file object.

        :type: File
        """
        par = self._parent
        while isinstance(par, Entity):
            par = par._parent
        return par

    @property
    def referring_objects(self):
        objs = []
        objs.extend(self.referring_blocks)
        objs.extend(self.referring_groups)
        objs.extend(self.referring_data_arrays)
        objs.extend(self.referring_tags)
        objs.extend(self.referring_multi_tags)
        objs.extend(self.referring_sources)
        return objs

    @property
    def referring_blocks(self):
        f = self.file
        return list(blk for blk in f.blocks
                    if blk.metadata is not None and blk.metadata.id == self.id)

    @property
    def referring_groups(self):
        f = self.file
        groups = []
        for blk in f.blocks:
            groups.extend(grp for grp in blk.groups
                          if (grp.metadata is not None and
                              grp.metadata.id == self.id))
        return groups

    @property
    def referring_data_arrays(self):
        f = self.file
        data_arrays = []
        for blk in f.blocks:
            data_arrays.extend(da for da in blk.data_arrays
                               if (da.metadata is not None and
                                   da.metadata.id == self.id))
        return data_arrays

    @property
    def referring_tags(self):
        f = self.file
        tags = []
        for blk in f.blocks:
            tags.extend(tg for tg in blk.tags
                        if (tg.metadata is not None and
                            tg.metadata.id == self.id))
        return tags

    @property
    def referring_multi_tags(self):
        f = self.file
        multi_tags = []
        for blk in f.blocks:
            multi_tags.extend(mt for mt in blk.multi_tags
                              if (mt.metadata is not None and
                                  mt.metadata.id == self.id))
        return multi_tags

    @property
    def referring_sources(self):
        f = self.file
        sources = []
        for blk in f.blocks:
            sources.extend(src for src in blk.sources
                           if (src.metadata is not None and
                               src.metadata.id == self.id))
        return sources

    def find_sections(self, filtr=lambda _: True, limit=None):
        """
        Get all child sections recursively.
        This method traverses the trees of all sections. The traversal is
        accomplished via breadth first and can be limited in depth.
        On each node or section a filter is applied.
        If the filter returns true the respective section will be added to the
        result list.  By default a filter is used that accepts all sections.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sections.
        :rtype: list of Section
        """
        if limit is None:
            limit = maxint
        return finders._find_sections(self, filtr, limit)

    def find_related(self, filtr=lambda _: True):
        """
        Get all related sections of this section.

        The result can be filtered. On each related section a filter is
        applied.  If the filter returns true the respective section will be
        added to the result list. By default a filter is used that accepts all
        sections.

        :param filtr: A filter function
        :type filtr:  function

        :returns: A list containing the matching related sections.
        :rtype: list of Section
        """
        result = []
        if self.parent is not None:
            result = finders._find_sections(self.parent, filtr, 1)
        if self in result:
            del result[result.index(self)]
        result += finders._find_sections(self, filtr, 1)
        return result

    @property
    def sections(self):
        """
        A property providing all child Sections of a Section. Child sections
        can be accessed by name, index, or id. Sections can also be deleted:
        if a Section is deleted, all its properties and child Sections are
        removed from the file too. Adding new Sections is achieved using the
        create_section method.
        This is a read-only attribute.

        :type: Container of Section
        """
        if self._sections is None:
            self._sections = SectionContainer("sections", self, Section)
        return self._sections

    def __len__(self):
        return len(self.props)

    def __getitem__(self, key):

        if key not in self.props and key in self.sections:
            return self.sections[key]

        prop = self.props[key]
        values = list(prop.values)
        if len(values) == 1:
            values = values[0]
        return values

    def __delitem__(self, key):
        del self.props[key]

    def __setitem__(self, key, data):

        if isinstance(data, S):
            data.section = self.create_section(key, data.section_type)
            return

        if not isinstance(data, list):
            data = [data]

        if key not in self.props:
            self.create_property(key, data)
        else:
            prop = self.props[key]
            prop.values = data

    def __iter__(self):
        for name, item in self.items():
            yield item

    def items(self):
        for p in self.props:
            yield (p.name, p)
        for s in self.sections:
            yield (s.name, s)

    def __contains__(self, key):
        return key in self.props or key in self.sections

    @property
    def props(self):
        """
        A property containing all Property entities associated with the
        section. Properties can be accessed by name, index, or id.
        Properties can be deleted from the list. Adding new Properties is done
        using the create_property method.
        This is a read-only attribute.

        :type: Container of Property
        """
        if self._properties is None:
            self._properties = Container("properties", self, Property)
        return self._properties

    def pprint(self, indent=2, max_depth=1, max_length=80, current_depth=0):
        """
        Pretty print method.

        :param indent: The number of indentation spaces per recursion
        :type indent:  int
        :param max_depth: The maximum times of recursion
        :type max_depth:  int
        :param max_length: The maximum length of each line of output
        :type max_length:  int
        :param current_depth: The current times of recursion
        :type current_depth:  int
        """

        spaces = " " * (current_depth * indent)
        sec_str = "{}{} [{}]".format(spaces, self.name, self.type)
        print(sec_str)
        for p in self.props:
            p.pprint(current_depth=current_depth, indent=indent,
                     max_length=max_length)
        if max_depth == -1 or current_depth < max_depth:
            for s in self.sections:
                s.pprint(current_depth=current_depth+1, max_depth=max_depth,
                         indent=indent, max_length=max_length)
        elif max_depth == current_depth:
            child_sec_indent = spaces + " " * indent
            more_indent = spaces + " " * (current_depth + 2 * indent)
            for s in self.sections:
                print("{} {} [{}]\n{}[...]".format(child_sec_indent,
                                                   s.name, s.type,
                                                   more_indent))

    @staticmethod
    def _change_id(_, grp):
        if "entity_id" in grp.attrs:
            id_ = util.create_id()
            grp.attrs.modify("entity_id", np.string_(id_))



# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
import os
import gc
import numpy as np
from warnings import warn

try:
    from sys import maxint
except ImportError:
    from sys import maxsize as maxint
import h5py

from .hdf5.h5group import H5Group
from .block import Block
from .section import Section
from .container import Container, SectionContainer
from . import util
from .exceptions import InvalidFile, DuplicateName
from .util import find as finders
from . import validator
from .compression import Compression


FILE_FORMAT = "nix"
HDF_FF_VERSION = (1, 1, 1)


def can_write(nixfile):
    filever = nixfile.version
    if len(filever) != 3:
        raise RuntimeError("Invalid version specified in file.")
    if HDF_FF_VERSION == filever:
        return True
    else:
        return False


def can_read(nixfile):
    filever = nixfile.version
    if len(filever) != 3:
        raise RuntimeError("Invalid version specified in file.")
    vx, vy, vz = HDF_FF_VERSION
    fx, fy, fz = filever
    if vx == fx and vy >= fy:
        return True
    else:
        return False


class FileMode(object):
    ReadOnly = 'r'
    ReadWrite = 'a'
    Overwrite = 'w'


def map_file_mode(mode):
    if mode == FileMode.ReadOnly:
        return h5py.h5f.ACC_RDONLY
    elif mode == FileMode.ReadWrite:
        return h5py.h5f.ACC_RDWR
    elif mode == FileMode.Overwrite:
        return h5py.h5f.ACC_TRUNC
    else:
        raise ValueError("Invalid file mode specified.")


def make_fapl():
    return h5py.h5p.create(h5py.h5p.FILE_ACCESS)


def make_fcpl():
    fcpl = h5py.h5p.create(h5py.h5p.FILE_CREATE)
    flags = h5py.h5p.CRT_ORDER_TRACKED | h5py.h5p.CRT_ORDER_INDEXED
    fcpl.set_link_creation_order(flags)
    return fcpl


class File(object):

    def __init__(self, path, mode=FileMode.ReadWrite,
                 compression=Compression.Auto, auto_update_time=False):
        """
        Open a NIX file, or create it if it does not exist.

        :param path: Path to file
        :param mode: FileMode ReadOnly, ReadWrite, or Overwrite.
                    (default: ReadWrite)
        :param compression: No, DeflateNormal, Auto (default: Auto)
        :return: nixio.File object
        """
        try:
            path = path.encode("utf-8")
        except (UnicodeError, LookupError):
            pass

        if not os.path.exists(path) and mode == FileMode.ReadOnly:
            raise RuntimeError(
                "Cannot open non-existent file in ReadOnly mode!"
            )

        new = False
        if not os.path.exists(path) or mode == FileMode.Overwrite:
            mode = FileMode.Overwrite
            h5mode = map_file_mode(mode)
            fid = h5py.h5f.create(path, flags=h5mode, fapl=make_fapl(),
                                  fcpl=make_fcpl())
            new = True
        else:
            h5mode = map_file_mode(mode)
            fid = h5py.h5f.open(path, flags=h5mode, fapl=make_fapl())

        self._h5file = h5py.File(fid)
        self._root = H5Group(self._h5file, "/", create=True)
        self._h5group = self._root  # to match behaviour of other objects
        self._time_auto_update = True
        if new:
            self._create_header()
        self._check_header(mode)
        self.mode = mode
        self._data = self._root.open_group("data", create=True)
        self.metadata = self._root.open_group("metadata", create=True)
        if "created_at" not in self._h5file.attrs:
            self.force_created_at()
        if "updated_at" not in self._h5file.attrs:
            self.force_updated_at()
        self.time_auto_update = auto_update_time
        if compression == Compression.Auto:
            compression = Compression.No
        self._compr = compression
        # make container props but don't initialise
        self._blocks = None
        self._sections = None

    @classmethod
    def open(cls, path, mode=FileMode.ReadWrite, compression=Compression.Auto,
             backend=None,  auto_update_time=False):
        if backend is not None:
            warn("Backend selection is deprecated. Ignoring value.")
        return cls(path, mode, compression, auto_update_time)

    def _create_header(self):
        self.format = FILE_FORMAT
        self.version = HDF_FF_VERSION

    def _check_header(self, mode):
        if self.format != FILE_FORMAT:
            raise InvalidFile

        if mode == FileMode.ReadWrite:
            if not can_write(self):
                raise RuntimeError("Cannot open file for writing. "
                                   "Incompatible version.")
        elif mode == FileMode.ReadOnly:
            if not can_read(self):
                raise RuntimeError("Cannot open file. "
                                   "Incompatible version.")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def version(self):
        """
        The file format version.

        :type: tuple
        """
        return tuple(self._root.get_attr("version"))

    @version.setter
    def version(self, v):
        util.check_attr_type(v, tuple)
        for part in v:
            util.check_attr_type(part, int)
        # convert to np.int32 since py3 defaults to 64
        v = np.array(v, dtype=np.int32)
        self._root.set_attr("version", v)
        if self.time_auto_update:
            self.force_updated_at()

    @property
    def format(self):
        """
        The format of the file. This read only property should always have the
        value 'nix'.

        :type: str
        """
        return self._root.get_attr("format")

    @format.setter
    def format(self, f):
        util.check_attr_type(f, str)
        self._root.set_attr("format", f.encode("ascii"))
        if self.time_auto_update:
            self.force_updated_at()

    @property
    def time_auto_update(self):
        """
        A user defined flag which decided if time should always be updated
        when properties are changed.

        :type: bool
        """
        return self._time_auto_update

    @time_auto_update.setter
    def time_auto_update(self, auto_update_flag):
        self._time_auto_update = auto_update_flag

    @property
    def created_at(self):
        """
        The creation time of the file. This is a read-only property.
        Use `force_created_at` in order to change the creation time.

        :rtype: int
        """
        return util.str_to_time(self._h5file.attrs["created_at"])

    def force_created_at(self, t=None):
        """
        Sets the creation time `created_at` to the given time
        (default: current time).

        :param t: The time to set
        :type t: int
        """
        if t is None:
            t = util.now_int()
        else:
            util.check_attr_type(t, int)
        self._h5file.attrs["created_at"] = util.time_to_str(t)

    @property
    def updated_at(self):
        """
        The time of the last update of the file. This is a read-only
        property. Use `force_updated_at` in order to change the update
        time.

        :rtype: int
        """
        return util.str_to_time(self._h5file.attrs["updated_at"])

    def force_updated_at(self, t=None):
        """
        Sets the update time `updated_at` to the given time.
        (default: current time)

        :param t: The time to set (default: now)
        :type t: int
        """
        if t is None:
            t = util.now_int()
        else:
            util.check_attr_type(t, int)
        self._h5file.attrs["updated_at"] = util.time_to_str(t)

    def is_open(self):
        """
        Checks whether a file is open or closed.

        :returns: True if the file is open, False otherwise.
        :rtype: bool
        """
        try:
            self._h5file.mode
            return True
        except ValueError:
            return False

    def validate(self):
        return validator.check_file(self)

    def pprint(self, indent=2, max_length=120, extra=True, max_depth=3):
        """
        Pretty Printing the Data and MetaData Tree of the whole File

        :param indent: The length of one indentation space
        :type indent: int
        :param max_length: Maximum length of each line of output
        :type max_length: int
        :param extra: True to print extra information of Entities
        :type extra: bool
        :param max_depth: Maximum recursion being printed in MetaData tree
        :type max_depth: int
        """
        print("File: name = {}".format(self._h5group.group.file.filename))
        if self.blocks:
            for blk in self.blocks:
                blk.pprint(indent=indent,
                           max_length=max_length, extra=extra, start_depth=1)
        if self.sections:
            for sec in self.sections:
                sec.pprint(indent=indent, max_depth=max_depth,
                           max_length=max_length, current_depth=1)

    # TODO: if same file, set_attr("entity_id", id_)

    def copy_section(self, obj, children=True, keep_id=True, name=""):
        """
        Copy a section to the file.

        :param obj: The Section to be copied
        :type obj: Section
        :param children: Specify if the copy should be recursive
        :type children: bool
        :param keep_id: Specify if the id should be kept
        :type keep_id: bool
        :param name: Name of copied section, Default is name of source section
        :type name: str

        :returns: The copied section
        :rtype: Section
        """
        if not isinstance(obj, Section):
            raise TypeError("Object to be copied is not a Section")

        if obj._sec_parent:
            src = "{}/{}".format("sections", obj.name)
        else:
            src = "{}/{}".format("metadata", obj.name)
        clsname = "metadata"
        if not name:
            name = str(obj.name)
        sec = self._h5group.open_group("sections", True)
        if name in sec:
            raise NameError("Name already exist. Possible solution is to "
                            "provide a new name when copying destination "
                            "is the same as the source parent")
        obj._parent._h5group.copy(source=src, dest=self._h5group,
                                  name=name, cls=clsname,
                                  shallow=not children, keep_id=keep_id)

        if not children:
            for p in obj.props:
                self.sections[obj.name].create_property(copy_from=p,
                                                        keep_copy_id=keep_id)

        return self.sections[obj.name]

    def flush(self):
        self._h5file.flush()

    def close(self):
        """
        Closes an open file.
        """
        gc.collect()  # should handle refs better instead of calling collect()
        # Flush is probably unnecessary
        self._h5file.flush()
        self._h5file.close()

    # Block
    def create_block(self, name="", type_="", compression=Compression.Auto,
                     copy_from=None, keep_copy_id=True):
        """
        Create a new block inside the file.

        :param name: The name of the block to create.
        :type name: str
        :param type_: The type of the block.
        :type type_: str
        :param compression: No, DeflateNormal, Auto (default: Auto)
        :param copy_from: The Block to be copied, None in normal mode
        :type copy_from: Block
        :param keep_copy_id: Specify if the id should be copied in copy mode
        :type keep_copy_id: bool

        :returns: The newly created block.
        :rtype: Block
        """
        if copy_from:
            if not isinstance(copy_from, Block):
                raise TypeError("Object to be copied is not a Block")
            clsname = "data"
            src = "{}/{}".format(clsname, copy_from.name)
            if not name:
                name = str(copy_from.name)
            if name in self._data:
                raise NameError("Name already exist. Possible solution is to "
                                "provide a new name when copying destination "
                                "is the same as the source parent")
            b = copy_from._parent._h5group.copy(source=src, dest=self._h5group,
                                                name=name,
                                                cls=clsname,
                                                keep_id=keep_copy_id)
            id_ = b.attrs["entity_id"]
            return self.blocks[id_]

        if name in self._data:
            raise ValueError("Block with the given name already exists!")
        if compression == Compression.Auto:
            compression = self._compr
        block = Block._create_new(self, self._data, name, type_, compression)
        return block

    # Section
    def create_section(self, name, type_="undefined", oid=None):
        """
        Create a new metadata section inside the file.

        :param name: The name of the section to create.
        :type name: str
        :param type_: The type of the section.
        :type type_: str
        :param oid: object id, UUID string as specified in RFC 4122. If no id
                    is provided, an id will be generated and assigned.
        :type oid: str

        :returns: The newly created section.
        :rtype: Section
        """
        if name in self.metadata:
            raise DuplicateName("create_section")
        sec = Section._create_new(self, self.metadata, name, type_, oid)
        return sec

    @property
    def blocks(self):
        """
        A property containing all blocks of a file. Blocks can be obtained by
        their name, id or index. Blocks can be deleted from the list, when a
        block is deleted all its content (data arrays, tags and sources) will
        be also deleted from the file. Adding new Block is done via the
        create_block method of File. This is a read-only attribute.
        """
        if self._blocks is None:
            self._blocks = Container("data", self, Block)
        return self._blocks

    def find_sections(self, filtr=lambda _: True, limit=None):
        """
        Get all sections and their child sections recursively.

        This method traverses the trees of all sections. The traversal is
        accomplished via breadth first and can be limited in depth. On each
        node or section a filter is applied. If the filter returns true the
        respective section will be added to the result list.
        By default a filter is used that accepts all sections.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sections.
        :rtype: list of Section
        """
        if limit is None:
            limit = maxint
        return finders._find_sections(self, filtr, limit)

    @property
    def sections(self):
        """
        A property containing all root Sections of a file. Specific root
        Sections can be obtained by their name, id or index. Sections can be
        deleted from this list. Notice: when a section is deleted all its child
        section and properties will be removed too. Adding a new Section is
        done via the crate_section method of File.
        This is a read-only property.
        """
        if self._sections is None:
            self._sections = SectionContainer("metadata", self, Section)
        return self._sections


# Copy File constructor docstring to File.open
File.open.__func__.__doc__ = File.__init__.__doc__



# -*- coding: utf-8 -*-
# Copyright © 2015, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from numbers import Integral, Real
from six import string_types

import numpy as np


BOOLS = (bool, np.bool_)


class DataType(object):
    UInt8 = np.uint8
    UInt16 = np.uint16
    UInt32 = np.uint32
    UInt64 = np.uint64
    Int8 = np.int8
    Int16 = np.int16
    Int32 = np.int32
    Int64 = np.int64
    Float = np.float32
    Double = np.double
    String = np.unicode_
    Bool = np.bool_

    @classmethod
    def get_dtype(cls, value):
        if isinstance(value, BOOLS):
            return cls.Bool
        elif isinstance(value, Integral):
            return cls.Int64
        elif isinstance(value, Real):
            return cls.Double
        elif isinstance(value, string_types):
            return cls.String
        else:
            raise ValueError("Unknown type for value {}".format(value))

    @classmethod
    def is_numeric_dtype(cls, dtype):
        return (dtype == cls.Int8 or
                dtype == cls.Int16 or
                dtype == cls.Int32 or
                dtype == cls.Int64 or
                dtype == cls.UInt8 or
                dtype == cls.UInt16 or
                dtype == cls.UInt32 or
                dtype == cls.UInt64 or
                dtype == cls.Float or
                dtype == cls.Double)



# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from numbers import Integral
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from .data_set import DataSet
from .exceptions import OutOfBounds, IncompatibleDimensions


class DataView(DataSet):

    def __init__(self, da, slices):
        if len(slices) != len(da.shape):
            # This is always checked by the calling function, but we repeat
            # the check here for future bug catching
            raise IncompatibleDimensions(
                "Number of dimensions for DataView does not match underlying "
                "DataArray: {} != {}".format(len(slices), len(da.shape)),
                "DataView"
            )

        if any(s.stop > e for s, e in zip(slices, da.data_extent)):
            raise OutOfBounds(
                "Trying to create DataView which is out of bounds of the "
                "underlying DataArray"
            )

        # Simplify all slices
        slices = tuple(slice(*sl.indices(dimlen))
                       for sl, dimlen in zip(slices, da.shape))

        self.array = da
        self._h5group = self.array._h5group
        self._slices = slices

    @property
    def data_extent(self):
        return tuple(s.stop - s.start for s in self._slices)

    @data_extent.setter
    def data_extent(self, v):
        raise AttributeError("can't set attribute")

    @property
    def data_type(self):
        return self.array.data_type

    def _write_data(self, data, sl=None):
        tsl = self._slices
        if sl:
            tsl = self._transform_coordinates(sl)
        super(DataView, self)._write_data(data, tsl)

    def _read_data(self, sl=None):
        tsl = self._slices
        if sl:
            tsl = self._transform_coordinates(sl)
        return super(DataView, self)._read_data(tsl)

    def _transform_coordinates(self, user_slices):
        """
        Takes a series (tuple) of slices or indices passed to the DataView and
        transforms them to the equivalent slices or indices for the underlying
        DataArray. Bounds checking is performed on the results to make sure it
        is not outside the DataView's range.

        Note: HDF5 hyperslabs don't support negative steps, so we catch it
        here to throw an error from NIX instead to shorten the stack trace (we
        use the same message).
        """
        oob = OutOfBounds("Trying to access data outside of range of DataView")

        def transform_slice(uslice, dvslice):
            """
            Single dimension transform function for slices.

            uslice: User provided slice for dimension
            dvslice: DataView slice for dimension
            """
            # Simplify uslice; DataView step is always 1
            dimlen = dvslice.stop - dvslice.start
            ustart, ustop, ustep = uslice.indices(dimlen)
            if ustop < 0:  # special case for None stop
                ustop = dimlen + ustop
            tslice = slice(dvslice.start+ustart, dvslice.start+ustop, ustep)
            if tslice.stop > dvslice.stop:
                raise oob

            if tslice.step < 0:
                raise ValueError("Step must be >= 1")

            return tslice

        dvslices = self._slices
        user_slices = self._expand_user_slices(user_slices)
        tslices = list()
        for uslice, dvslice in zip(user_slices, dvslices):
            if isinstance(uslice, Integral):
                if uslice < 0:
                    tslice = dvslice.stop + uslice
                else:
                    tslice = uslice + dvslice.start
                if tslice < dvslice.start or tslice >= dvslice.stop:
                    raise oob
            elif isinstance(uslice, slice):
                tslice = transform_slice(uslice, dvslice)
                if tslice.start < dvslice.start:
                    raise oob
                if tslice.stop > dvslice.stop:
                    raise oob
            else:
                raise TypeError("Data indices must be integers or slices, "
                                "not {}".format(type(uslice)))
            tslices.append(tslice)

        return tuple(tslices)

    def _expand_user_slices(self, user_slices):
        """
        Given the user-supplied slices or indices, expands Ellipses if
        necessary and returns the same objects in a tuple padded with
        slice(None) to match the dimensionality of the DataView.
        """
        if not isinstance(user_slices, Iterable):
            user_slices = (user_slices,)

        if user_slices.count(Ellipsis) > 1:
            raise IndexError(
                "an index can only have a single ellipsis ('...')"
            )
        elif user_slices.count(Ellipsis) == 1:
            # expand slices at Ellipsis index
            expidx = user_slices.index(Ellipsis)
            npad = len(self.data_extent) - len(user_slices) + 1
            padding = (slice(None),) * npad
            return user_slices[:expidx] + padding + user_slices[expidx+1:]

        # expand slices at the end
        npad = len(self.data_extent) - len(user_slices)
        padding = (slice(None),) * npad
        return user_slices + padding



# -*- coding: utf-8 -*-
# Copyright © 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from .util import *
from . import names
from . import units
